""" Program for validating that simulation tools are consistent with the BioSimulators standards

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-22
:Copyright: 2020, BioSimulators Team
:License: MIT
"""

from .combine.data_model import CombineArchiveContentFormat
from .config import get_config
from .sedml.data_model import ModelLanguage, OneStepSimulation, SteadyStateSimulation, UniformTimeCourseSimulation
from .utils.core import flatten_nested_list_of_strings
from .warnings import warn, BioSimulatorsWarning
import biosimulators_utils
import cement
import json
import shutil
import tempfile


class BaseController(cement.Controller):
    """ Base controller for command line application """

    class Meta:
        label = 'base'
        description = "Utilities for working with containerized biosimulation tools"
        help = "Utilities for working with containerized biosimulation tools"
        arguments = [
            (['-v', '--version'], dict(
                action='version',
                version=biosimulators_utils.__version__,
            )),
        ]

    @cement.ex(hide=True)
    def _default(self):
        self._parser.print_help()


BUILD_COMBINE_ARCHIVE_MODEL_LANGUAGES = [
    ModelLanguage.BNGL,
    ModelLanguage.CellML,
    # ModelLanguage.ZGINML,
    # ModelLanguage.LEMS,
    # ModelLanguage.NeuroML,
    ModelLanguage.RBA,
    ModelLanguage.SBML,
    ModelLanguage.Smoldyn,
    ModelLanguage.XPP,
]

VALIDATE_MODEL_LANGUAGES = [
    ModelLanguage.BNGL,
    ModelLanguage.CellML,
    # ModelLanguage.ZGINML,
    ModelLanguage.LEMS,
    ModelLanguage.NeuroML,
    ModelLanguage.RBA,
    ModelLanguage.SBML,
    ModelLanguage.Smoldyn,
    ModelLanguage.XPP,
]

BUILD_COMBINE_ARCHIVE_SIMULATION_TYPES = [
    OneStepSimulation,
    SteadyStateSimulation,
    UniformTimeCourseSimulation,
]


class BuildModelingProjectController(cement.Controller):
    """ Controller for building a COMBINE/OMEX archive with a SED-ML file for a model """

    class Meta:
        label = 'build-project'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = "Build a modeling project"
        description = "Build a COMBINE/OMEX archive with a SED-ML file for a model"
        arguments = [
            (
                ['model_language'],
                dict(
                    metavar='model-language',
                    type=str,
                    help='Model language ({}, or {})'.format(
                        ', '.join('`{}`'.format(lang.name) for lang in BUILD_COMBINE_ARCHIVE_MODEL_LANGUAGES[0:-1]),
                        ', '.join('`{}`'.format(lang.name) for lang in BUILD_COMBINE_ARCHIVE_MODEL_LANGUAGES[-1:]),
                    ),
                ),
            ),
            (
                ['model_filename'],
                dict(
                    metavar='model-filename',
                    type=str,
                    help='Path to model',
                ),
            ),
            (
                ['simulation_type'],
                dict(
                    metavar='simulation-type',
                    type=str,
                    help='Type of simulation ({}, or {})'.format(
                        ', '.join('`{}`'.format(type.__name__[0:-10]) for type in BUILD_COMBINE_ARCHIVE_SIMULATION_TYPES[0:-1]),
                        ', '.join('`{}`'.format(type.__name__[0:-10]) for type in BUILD_COMBINE_ARCHIVE_SIMULATION_TYPES[-1:]),
                    ),
                ),
            ),
            (
                ['archive_filename'],
                dict(
                    metavar='archive-filename',
                    type=str,
                    help='Path to save archive',
                ),
            ),
        ]

    @cement.ex(hide=True)
    def _default(self):
        from .sedml.model_utils import build_combine_archive_for_model

        args = self.app.pargs

        try:
            model_lang = ModelLanguage[args.model_language]
            assert model_lang in BUILD_COMBINE_ARCHIVE_MODEL_LANGUAGES
        except (KeyError, AssertionError):
            raise SystemExit('Model language must be one of {}, or {}, not `{}`.'.format(
                ', '.join('`{}`'.format(lang.name) for lang in BUILD_COMBINE_ARCHIVE_MODEL_LANGUAGES[0:-1]),
                ', '.join('`{}`'.format(lang.name) for lang in BUILD_COMBINE_ARCHIVE_MODEL_LANGUAGES[-1:]),
                args.model_language,
            ))

        model_filename = args.model_filename

        sim_type = None
        for a_sim_type in BUILD_COMBINE_ARCHIVE_SIMULATION_TYPES:
            if a_sim_type.__name__ == args.simulation_type + 'Simulation':
                sim_type = a_sim_type
                break
        if sim_type is None:
            raise SystemExit('Simulation type must be one of {}, or {}, not `{}`.'.format(
                ', '.join('`{}`'.format(type.__name__[0:-10]) for type in BUILD_COMBINE_ARCHIVE_SIMULATION_TYPES[0:-1]),
                ', '.join('`{}`'.format(type.__name__[0:-10]) for type in BUILD_COMBINE_ARCHIVE_SIMULATION_TYPES[-1:]),
                args.simulation_type,
            ))

        build_combine_archive_for_model(model_filename, model_lang, sim_type, args.archive_filename)


class ValidateModelController(cement.Controller):
    """ Controller for validating models (e.g., CellML, SBML files) """

    class Meta:
        label = 'validate-model'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = "Validate a model (e.g., CellML, SBML file)"
        description = "Validate a model (e.g., CellML, SBML file)"
        arguments = [
            (
                ['language'],
                dict(
                    type=str,
                    help='Model language ({}, or {})'.format(
                        ', '.join('`{}`'.format(lang.name) for lang in VALIDATE_MODEL_LANGUAGES[0:-1]),
                        ', '.join('`{}`'.format(lang.name) for lang in VALIDATE_MODEL_LANGUAGES[-1:]),
                    ),
                ),
            ),
            (
                ['filename'],
                dict(
                    type=str,
                    help='Path to model',
                ),
            ),
        ]

    @cement.ex(hide=True)
    def _default(self):
        import biosimulators_utils.sedml.validation

        args = self.app.pargs

        try:
            language = ModelLanguage[args.language]
            assert language in BUILD_COMBINE_ARCHIVE_MODEL_LANGUAGES
        except (KeyError, AssertionError):
            raise SystemExit('Model language must be one of {}, or {}, not `{}`.'.format(
                ', '.join('`{}`'.format(lang.name) for lang in BUILD_COMBINE_ARCHIVE_MODEL_LANGUAGES[0:-1]),
                ', '.join('`{}`'.format(lang.name) for lang in BUILD_COMBINE_ARCHIVE_MODEL_LANGUAGES[-1:]),
                args.language,
            ))

        filename = args.filename

        errors, warnings, _ = biosimulators_utils.sedml.validation.validate_model_with_language(filename, language)

        if warnings:
            msg = 'The model file `{}` may be invalid.\n  {}'.format(
                filename, flatten_nested_list_of_strings(warnings).replace('\n', '\n  '))
            warn(msg, BioSimulatorsWarning)

        if errors:
            msg = 'The model file `{}` is invalid.\n  {}'.format(
                filename, flatten_nested_list_of_strings(errors).replace('\n', '\n  '))
            raise SystemExit(msg)


class ValidateSimulationController(cement.Controller):
    """ Controller for validating simulation experiments (SED-ML files) """

    class Meta:
        label = 'validate-simulation'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = "Validate a simulation experiment (SED-ML file)"
        description = "Validate a simulation experiment (SED-ML file)"
        arguments = [
            (
                ['filename'],
                dict(
                    type=str,
                    help='Path to a SED-ML file',
                ),
            ),
        ]

    @cement.ex(hide=True)
    def _default(self):
        import biosimulators_utils.sedml.io

        args = self.app.pargs

        reader = biosimulators_utils.sedml.io.SedmlSimulationReader()
        try:
            reader.run(args.filename, validate_models_with_languages=False, validate_targets_with_model_sources=False)
        except Exception:
            if not reader.errors:
                raise

        if reader.warnings:
            msg = 'The SED-ML file `{}` may be invalid.\n  {}'.format(
                args.filename, flatten_nested_list_of_strings(reader.warnings).replace('\n', '\n  '))
            warn(msg, BioSimulatorsWarning)

        if reader.errors:
            msg = 'The SED-ML file `{}` is invalid.\n  {}'.format(
                args.filename, flatten_nested_list_of_strings(reader.errors).replace('\n', '\n  '))
            raise SystemExit(msg)


class ValidateMetadataController(cement.Controller):
    """ Controller for validating metadata (OMEX Metadata files) """

    class Meta:
        label = 'validate-metadata'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = "Validate metadata (OMEX Metadata file)"
        description = "Validate metadata (OMEX Metadata file)"
        arguments = [
            (
                ['filename'],
                dict(
                    type=str,
                    help='Path to an OMEX Metadata file',
                ),
            ),
        ]

    @cement.ex(hide=True)
    def _default(self):
        import biosimulators_utils.omex_meta.io

        args = self.app.pargs

        _, errors, warnings = biosimulators_utils.omex_meta.io.read_omex_meta_file(args.filename)

        if warnings:
            msg = 'The OMEX Metadata file `{}` may be invalid.\n  {}'.format(
                args.filename, flatten_nested_list_of_strings(warnings).replace('\n', '\n  '))
            warn(msg, BioSimulatorsWarning)

        if errors:
            msg = 'The OMEX Metadata file `{}` is invalid.\n  {}'.format(
                args.filename, flatten_nested_list_of_strings(errors).replace('\n', '\n  '))
            raise SystemExit(msg)


class ValidateModelingProjectController(cement.Controller):
    """ Controller for validating modeling projects (COMBINE/OMEX archives) """

    class Meta:
        label = 'validate-project'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = "Validate a model project (COMBINE/OMEX archive)"
        description = "Validate a modeling project (COMBINE/OMEX archive and its contents)"
        arguments = [
            (
                ['filename'],
                dict(
                    type=str,
                    help='Path to a COMBINE/OMEX archive',
                ),
            ),
        ]

    @cement.ex(hide=True)
    def _default(self):
        import biosimulators_utils.combine.io
        import biosimulators_utils.combine.utils
        import biosimulators_utils.combine.validation

        args = self.app.pargs

        archive_dirname = tempfile.mkdtemp()

        config = get_config()

        reader = biosimulators_utils.combine.io.CombineArchiveReader()
        try:
            archive = reader.run(args.filename, archive_dirname, config=config)
        except Exception as exception:
            shutil.rmtree(archive_dirname)
            raise SystemExit(str(exception))

        errors, warnings = biosimulators_utils.combine.validation.validate(
            archive, archive_dirname,
            formats_to_validate=list(CombineArchiveContentFormat.__members__.values()),
            config=config,
        )
        if warnings:
            msg = 'The COMBINE/OMEX archive may be invalid.\n  {}'.format(
                flatten_nested_list_of_strings(warnings).replace('\n', '\n  '))
            warn(msg, BioSimulatorsWarning)

        if errors:
            shutil.rmtree(archive_dirname)

            msg = '`{}` is not a valid COMBINE/OMEX archive.\n  {}'.format(
                args.filename, flatten_nested_list_of_strings(errors).replace('\n', '\n  '))
            raise SystemExit(msg)

        # print summary
        print(biosimulators_utils.combine.utils.get_summary_sedml_contents(archive, archive_dirname))

        # clean up
        shutil.rmtree(archive_dirname)


class ExecuteModelingProjectController(cement.Controller):
    """ Controller for using containerized simulation tools to execute modeling projects (COMBINE/OMEX archives) """

    class Meta:
        label = 'exec'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = "Execute a model project (COMBINE/OMEX archive)"
        description = "Use a containerized simulation tool to execute a modeling project (COMBINE/OMEX archive)"
        arguments = [
            (
                ['docker_image'],
                dict(
                    metavar='docker-image',
                    type=str,
                    help='Tag or URL for the Docker image of the simulation tools (e.g., ghcr.io/biosimulators/tellurium:latest)',
                ),
            ),
            (
                ['-i', '--archive'],
                dict(
                    type=str,
                    required=True,
                    help='Path to a COMBINE/OMEX file which contains one or more SED-ML-encoded simulation experiments',
                ),
            ),
            (
                ['-o', '--out-dir'],
                dict(
                    type=str,
                    default='.',
                    help='Directory to save outputs',
                ),
            ),
            (
                ['--do-not-pull-image'],
                dict(
                    action='store_true',
                    help="If set, do not pull Docker image",
                ),
            ),
            (
                ['--env'],
                dict(
                    type=str,
                    nargs='+',
                    default=[],
                    help="Key-value pairs of values of environment variables (e.g., `KEY=value`)",
                ),
            ),
            (
                ['--tmp-dir'],
                dict(
                    type=str,
                    default='/tmp',
                    help="Path to temporary directory in Docker image. Default: `/tmp`.",
                ),
            ),
            (
                ['--path-sep'],
                dict(
                    type=str,
                    default='/',
                    help="Separator for file paths in Docker image. Default: `/`.",
                ),
            ),
            (
                ['--user'],
                dict(
                    type=str,
                    default="_CURRENT_USER_",
                    help="Id of user to execute commands within the Docker container. Default: `_CURRENT_USER_`",
                ),
            ),
            (
                ['--no-tty'],
                dict(
                    action='store_true',
                    help="Do not allocate a pseudo-TTY",
                ),
            ),
            (
                ['--keep-container'],
                dict(
                    action='store_true',
                    help="Keep Docker container (do not automatically remove the container when the archive is finished executing)",
                ),
            ),
        ]

    @cement.ex(hide=True)
    def _default(self):
        import biosimulators_utils.simulator.exec

        args = self.app.pargs

        try:
            env = {}
            for key_value in args.env:
                key, sep, value = key_value.partition('=')
                if sep != '=':
                    raise ValueError('Environment variables must be pairs of keys and values (e.g., KEY=value)')
                env[key] = value

            biosimulators_utils.simulator.exec.exec_sedml_docs_in_archive_with_containerized_simulator(
                args.archive, args.out_dir, args.docker_image,
                docker_image_temp_dir=args.tmp_dir, docker_image_path_sep=args.path_sep,
                environment=env, pull_docker_image=not args.do_not_pull_image,
                user_to_exec_within_container=args.user, allocate_tty=not args.no_tty, remove_docker_container=not args.keep_container)

        except Exception as exception:
            raise SystemExit(str(exception))


class ConvertController(cement.Controller):
    """ Controller for converting among formats """

    class Meta:
        label = 'convert'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = "Convert files among formats"
        description = "Convert files among formats"
        arguments = []

    @cement.ex(hide=True)
    def _default(self):
        self._parser.print_help()


class ConvertEscherController(cement.Controller):
    """ Controller for converting Escher maps to Vega data visualizations """

    class Meta:
        label = 'escher-to-vega'
        stacked_on = 'convert'
        stacked_type = 'nested'
        help = "Convert an Escher map to Vega"
        description = "Convert an Escher map of a metabolic network to the Vega data visualization format"
        arguments = [
            (
                ['--data-sedml'],
                dict(
                    type=str,
                    help=(
                        'Id of a report in a SED-ML file and the location of the SED-ML file in its parent '
                        'COMBINE archive which can record the predicted flux of each reaction '
                        '(e.g., `path/to/simulation.sedml/Table_1`)'
                    ),
                    default=None,
                ),
            ),
            (
                ['--data-file'],
                dict(
                    type=str,
                    help=(
                        'Path to a JSON file which represents a list of dictionaries, each which have two keys '
                        '`label` and `values` whose values are the ids of reactions and arrays of length 1 of '
                        'their predicted fluxes'
                    ),
                    default=None,
                ),
            ),
            (
                ['--data-url'],
                dict(
                    type=str,
                    help=(
                        'URL for a JSON file which represents a list of dictionaries, each which have two keys '
                        '`label` and `values` whose values are the ids of reactions and arrays of length 1 of '
                        'their predicted fluxes'
                    ),
                    default=None,
                ),
            ),
            (
                ['escher_file'],
                dict(
                    metavar='escher-file',
                    type=str,
                    help='Path to the configuration of an Escher map',
                ),
            ),
            (
                ['vega_file'],
                dict(
                    metavar='vega-file',
                    type=str,
                    help='Path to save the diagram in Vega format',
                ),
            ),
        ]

    @cement.ex(hide=True)
    def _default(self):
        from biosimulators_utils.viz.vega.escher.core import escher_to_vega
        _convert_diagram(self.app.pargs, 'escher_file', escher_to_vega)


class ConvertGinmlController(cement.Controller):
    """ Controller for converting GINML activity flow diagrams to Vega data visualizations """

    class Meta:
        label = 'ginml-to-vega'
        stacked_on = 'convert'
        stacked_type = 'nested'
        help = "Convert a GINML activity flow diagram to Vega"
        description = "Convert a GINML activity flow diagram to the Vega data visualization format"
        arguments = [
            (
                ['--data-sedml'],
                dict(
                    action='store_true',
                    help=(
                        'Set to indicate that the predicted value of each node should be read from the reports of '
                        'the SED-ML files in a COMBINE archive'
                    ),
                ),
            ),
            (
                ['--data-file'],
                dict(
                    type=str,
                    help=(
                        'Path to a JSON file which represents a list of dictionaries, each which have two keys '
                        '`label` and `values` whose values are the ids of nodes and arrays of their predicted values'
                    ),
                    default=None,
                ),
            ),
            (
                ['--data-url'],
                dict(
                    type=str,
                    help=(
                        'URL for a JSON file which represents a list of dictionaries, each which have two keys '
                        '`label` and `values` whose values are the ids of nodes and arrays of their predicted values'
                    ),
                    default=None,
                ),
            ),
            (
                ['ginml_file'],
                dict(
                    metavar='ginml-file',
                    type=str,
                    help='Path to the GINML file',
                ),
            ),
            (
                ['vega_file'],
                dict(
                    metavar='vega-file',
                    type=str,
                    help='Path to save the diagram in Vega format',
                ),
            ),
        ]

    @cement.ex(hide=True)
    def _default(self):
        from biosimulators_utils.viz.vega.ginml.core import ginml_to_vega
        _convert_diagram(self.app.pargs, 'ginml_file', ginml_to_vega)


class ConvertSbgnController(cement.Controller):
    """ Controller for converting SBGN process description maps to Vega data visualizations """

    class Meta:
        label = 'sbgn-to-vega'
        stacked_on = 'convert'
        stacked_type = 'nested'
        help = "Convert a SBGN process description map to Vega"
        description = "Convert a SBGN process description map to the Vega data visualization format"
        arguments = [
            (
                ['--data-sedml'],
                dict(
                    type=str,
                    help=(
                        'Id of a report in a SED-ML file and the location of the SED-ML file in its parent '
                        'COMBINE archive which can record the predicted concentration of each glyph '
                        '(e.g., `path/to/simulation.sedml/Table_1`)'
                    ),
                    default=None,
                ),
            ),
            (
                ['--data-file'],
                dict(
                    type=str,
                    help=(
                        'Path to a JSON file which represents a list of dictionaries, each which have two keys '
                        '`label` and `values` whose values are the labels of glyphs and arrays of their predicted '
                        'concentrations'
                    ),
                    default=None,
                ),
            ),
            (
                ['--data-url'],
                dict(
                    type=str,
                    help=(
                        'URL for a JSON file which represents a list of dictionaries, each which have two keys '
                        '`label` and `values` whose values are the labels of glyphs and arrays of their predicted '
                        'concentrations'
                    ),
                    default=None,
                ),
            ),
            (
                ['sbgn_file'],
                dict(
                    metavar='sbgn-file',
                    type=str,
                    help='Path to the SBGN file',
                ),
            ),
            (
                ['vega_file'],
                dict(
                    metavar='vega-file',
                    type=str,
                    help='Path to save the diagram in Vega format',
                ),
            ),
        ]

    @cement.ex(hide=True)
    def _default(self):
        from biosimulators_utils.viz.vega.sbgn.core import sbgn_pd_map_to_vega
        _convert_diagram(self.app.pargs, 'sbgn_file', sbgn_pd_map_to_vega)


def _convert_diagram(args, diagram_file_attr, converter):
    n_data_args = (
        (args.data_sedml not in [None, False]) +
        (args.data_file is not None) +
        (args.data_url is not None)
    )

    if n_data_args == 0:
        raise SystemExit('One of --data-sedml, --data-file, or --data-url must be used')
    elif n_data_args > 1:
        raise SystemExit('Only one of --data-sedml, --data-file, or --data-url can be used')

    if args.data_sedml:
        if isinstance(args.data_sedml, bool):
            data_set = {'sedmlUri': []}
        else:
            sedDocumentlocation, _, reportId = args.data_sedml.rpartition('/')
            data_set = {'sedmlUri': [sedDocumentlocation, reportId]}

    elif args.data_file:
        with open(args.data_file, 'rb') as file:
            data_set = {'values': json.load(file)}

    else:
        data_set = {'url': args.data_url}

    try:
        converter(data_set, getattr(args, diagram_file_attr), args.vega_file)
    except Exception as exception:
        raise SystemExit(str(exception))


class App(cement.App):
    """ Command line application """
    class Meta:
        label = 'biosimulators-utils'
        base_controller = 'base'
        handlers = [
            BaseController,
            BuildModelingProjectController,
            ValidateModelController,
            ValidateSimulationController,
            ValidateMetadataController,
            ValidateModelingProjectController,
            ExecuteModelingProjectController,
            ConvertController,
            ConvertEscherController,
            ConvertGinmlController,
            ConvertSbgnController,
        ]


def main():
    with App() as app:
        app.run()



""" Data model for working with OMEX Metadata

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-06-23
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import enum
import rdflib.term  # noqa: F401

__all__ = [
    'Triple',
    'OmexMetadataInputFormat',
    'OmexMetadataOutputFormat',
    'OmexMetadataSchema',
    'BIOSIMULATIONS_ROOT_URI_FORMAT',
    'BIOSIMULATIONS_ROOT_URI_PATTERN',
    'BIOSIMULATIONS_PREDICATE_TYPES',
    'BIOSIMULATIONS_THUMBNAIL_FORMATS',
    'BIOSIMULATIONS_NAMESPACE_PREFIX_MAP',
    'BIOSIMULATIONS_THUMBNAIL_FORMATS',
]


class Triple(object):
    """ An RDF triple

    Attributes:
        subject (:obj:`rdflib.term.BNode`, :obj:`rdflib.term.Literal`, or :obj:`rdflib.term.URIRef`): subject
        predicate (:obj:`rdflib.term.BNode`, :obj:`rdflib.term.Literal`, or :obj:`rdflib.term.URIRef`): predict
        object (:obj:`rdflib.term.BNode`, :obj:`rdflib.term.Literal`, or :obj:`rdflib.term.URIRef`): object
    """

    def __init__(self, subject, predicate, object):
        """
        Args:
            subject (:obj:`rdflib.term.BNode`, :obj:`rdflib.term.Literal`, or :obj:`rdflib.term.URIRef`): subject
            predicate (:obj:`rdflib.term.BNode`, :obj:`rdflib.term.Literal`, or :obj:`rdflib.term.URIRef`): predict
            object (:obj:`rdflib.term.BNode`, :obj:`rdflib.term.Literal`, or :obj:`rdflib.term.URIRef`): object
        """
        self.subject = subject
        self.predicate = predicate
        self.object = object


class OmexMetadataInputFormat(str, enum.Enum):
    """ An format for reading an OMEX Metadata file """
    ntriples = 'ntriples'
    nquads = 'nquads'
    rdfa = 'rdfa'
    rdfxml = 'rdfxml'
    turtle = 'turtle'


class OmexMetadataOutputFormat(str, enum.Enum):
    """ An format for writing an OMEX Metadata file """
    dot = 'dot'
    json = 'json'
    json_triples = 'json-triples'
    html = 'html'
    ntriples = 'ntriples'
    nquads = 'nquads'
    rdfxml = 'rdfxml'
    rdfxml_abbrev = 'rdfxml-abbrev'
    rdfxml_xmp = 'rdfxml-xmp'
    turtle = 'turtle'


class OmexMetadataSchema(str, enum.Enum):
    """ Schema for OMEX Metadata documents """
    biosimulations = 'BioSimulations'
    rdf_triples = 'rdf_triples'


BIOSIMULATIONS_ROOT_URI_FORMAT = 'http://omex-library.org/{}.omex'
BIOSIMULATIONS_ROOT_URI_PATTERN = r'^(http://omex-library\.org/.*?\.omex)(/|$)'

BIOSIMULATIONS_PREDICATE_TYPES = {
    'http://dublincore.org/specifications/dublin-core/dcmi-terms/title': {
        'namespace': {
            'prefix': 'dc',
            'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/',
        },
        'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/title',
        'attribute': 'title',
        'label': 'title',
        'multiple_allowed': False,
        'has_uri': False,
        'has_label': True,
        'required': True,
    },
    'http://dublincore.org/specifications/dublin-core/dcmi-terms/abstract': {
        'namespace': {
            'prefix': 'dc',
            'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/',
        },
        'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/abstract',
        'attribute': 'abstract',
        'label': 'Abstract',
        'multiple_allowed': False,
        'has_uri': False,
        'has_label': True,
        'required': False,
    },
    'http://prismstandard.org/namespaces/basic/2.0/keyword': {
        'namespace': {
            'prefix': 'prism',
            'uri': 'http://prismstandard.org/namespaces/basic/2.0/',
        },
        'uri': 'http://prismstandard.org/namespaces/basic/2.0/keyword',
        'attribute': 'keywords',
        'label': 'Keywords',
        'multiple_allowed': True,
        'has_uri': False,
        'has_label': True,
        'required': False,
    },
    'http://dublincore.org/specifications/dublin-core/dcmi-terms/description': {
        'namespace': {
            'prefix': 'dc',
            'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/',
        },
        'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/description',
        'attribute': 'description',
        'label': 'Description',
        'multiple_allowed': False,
        'has_uri': False,
        'has_label': True,
        'required': False,
    },
    'http://biomodels.net/biology-qualifiers/hasTaxon': {
        'namespace': {
            'prefix': 'bqbiol',
            'uri': 'http://biomodels.net/biology-qualifiers/',
        },
        'uri': 'http://biomodels.net/biology-qualifiers/hasTaxon',
        'attribute': 'taxa',
        'label': 'Taxa',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': True,
        'required': False,
    },
    'http://biomodels.net/biology-qualifiers/encodes': {
        'namespace': {
            'prefix': 'bqbiol',
            'uri': 'http://biomodels.net/biology-qualifiers/',
        },
        'uri': 'http://biomodels.net/biology-qualifiers/encodes',
        'attribute': 'encodes',
        'label': 'Encodes',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': True,
        'required': False,
    },
    'http://www.collex.org/schema#thumbnail': {
        'namespace': {
            'prefix': 'collex',
            'uri': 'http://www.collex.org/schema#',
        },
        'uri': 'http://www.collex.org/schema#thumbnail',
        'attribute': 'thumbnails',
        'label': 'Thumbnails',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': False,
        'required': False,
    },
    'http://dublincore.org/specifications/dublin-core/dcmi-terms/source': {
        'namespace': {
            'prefix': 'dc',
            'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/',
        },
        'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/source',
        'attribute': 'sources',
        'label': 'Sources',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': True,
        'required': False,
    },
    'http://biomodels.net/model-qualifiers/isDerivedFrom': {
        'namespace': {
            'prefix': 'bqmodel',
            'uri': 'http://biomodels.net/model-qualifiers/',
        },
        'uri': 'http://biomodels.net/model-qualifiers/isDerivedFrom',
        'attribute': 'predecessors',
        'label': 'Predecessors',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': True,
        'required': False,
    },
    'http://purl.org/spar/scoro/successor': {
        'namespace': {
            'prefix': 'scoro',
            'uri': 'http://purl.org/spar/scoro/',
        },
        'uri': 'http://purl.org/spar/scoro/successor',
        'attribute': 'successors',
        'label': 'Successors',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': True,
        'required': False,
    },
    'http://www.w3.org/2000/01/rdf-schema#seeAlso': {
        'namespace': {
            'prefix': 'rdfs',
            'uri': 'http://www.w3.org/2000/01/rdf-schema#',
        },
        'uri': 'http://www.w3.org/2000/01/rdf-schema#seeAlso',
        'attribute': 'see_also',
        'label': 'See also',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': True,
        'required': False,
    },
    'http://dublincore.org/specifications/dublin-core/dcmi-terms/creator': {
        'namespace': {
            'prefix': 'dc',
            'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/',
        },
        'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/creator',
        'attribute': 'creators',
        'label': 'Creators',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': True,
        'required': True,
    },
    'http://dublincore.org/specifications/dublin-core/dcmi-terms/contributor': {
        'namespace': {
            'prefix': 'dc',
            'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/',
        },
        'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/contributor',
        'attribute': 'contributors',
        'label': 'Contributors',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': True,
        'required': False,
    },
    'http://biomodels.net/model-qualifiers/is': {
        'namespace': {
            'prefix': 'bqmodel',
            'uri': 'http://biomodels.net/model-qualifiers/',
        },
        'uri': 'http://biomodels.net/model-qualifiers/is',
        'attribute': 'identifiers',
        'label': 'Identifiers',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': True,
        'required': False,
    },
    'http://biomodels.net/model-qualifiers/isDescribedBy': {
        'namespace': {
            'prefix': 'bqmodel',
            'uri': 'http://biomodels.net/model-qualifiers/',
        },
        'uri': 'http://biomodels.net/model-qualifiers/isDescribedBy',
        'attribute': 'citations',
        'label': 'Citations',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': True,
        'required': False,
    },
    'http://dublincore.org/specifications/dublin-core/dcmi-terms/license': {
        'namespace': {
            'prefix': 'dc',
            'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/',
        },
        'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/license',
        'attribute': 'license',
        'label': 'License',
        'multiple_allowed': False,
        'has_uri': True,
        'has_label': True,
        'required': True
    },
    'http://purl.org/spar/scoro/funder': {
        'namespace': {
            'prefix': 'scoro',
            'uri': 'http://purl.org/spar/scoro/',
        },
        'uri': 'http://purl.org/spar/scoro/funder',
        'attribute': 'funders',
        'label': 'Funders',
        'multiple_allowed': True,
        'has_uri': True,
        'has_label': True,
        'required': False,
    },
    'http://dublincore.org/specifications/dublin-core/dcmi-terms/created': {
        'namespace': {
            'prefix': 'dc',
            'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/',
        },
        'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/created',
        'attribute': 'created',
        'label': 'Created',
        'multiple_allowed': False,
        'has_uri': False,
        'has_label': True,
        'required': False,
    },
    'http://dublincore.org/specifications/dublin-core/dcmi-terms/modified': {
        'namespace': {
            'prefix': 'dc',
            'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/',
        },
        'uri': 'http://dublincore.org/specifications/dublin-core/dcmi-terms/modified',
        'attribute': 'modified',
        'label': 'Modified',
        'multiple_allowed': True,
        'has_uri': False,
        'has_label': True,
        'required': False,
    },
}

BIOSIMULATIONS_NAMESPACE_PREFIX_MAP = {
    'http://biomodels.net/biology-qualifiers/': 'bqbiol',
    'http://biomodels.net/model-qualifiers/': 'biomodel',
    'http://dublincore.org/specifications/dublin-core/dcmi-terms/': 'dc',
    'http://prismstandard.org/namespaces/basic/2.0/': 'prism',
    'http://purl.org/dc/terms/': 'dcterms',
    'http://purl.org/spar/scoro/': 'scoro',
    'http://www.collex.org/schema#': 'collex',
    'http://www.w3.org/1999/02/22-rdf-syntax-ns#': 'rdf',
    'http://www.w3.org/2000/01/rdf-schema#': 'rdfs',
    'http://xmlns.com/foaf/0.1/': 'foaf',
}

BIOSIMULATIONS_THUMBNAIL_FORMATS = [
    'GIF',
    'JPEG',
    'PNG',
    'WEBP',
]



""" Data model for COMBINE/OMEX archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..utils.core import are_lists_equal, none_sorted
from ..data_model import Person  # noqa: F401
import abc
import datetime  # noqa: F401
import enum

__all__ = [
    'CombineArchiveBase',
    'CombineArchive',
    'CombineArchiveContent',
    'CombineArchiveContentFormat',
    'CombineArchiveContentFormatPattern',
]


class CombineArchiveBase(abc.ABC):
    """ A COMBINE/OMEX archive

    Attributes:
        description (:obj:`str`): description
        authors (:obj:`list` of :obj:`Person`): authors
        created (:obj:`datetime.datetime`): created date
        updated (:obj:`datetime.datetime`): updated date
    """
    pass


class CombineArchive(CombineArchiveBase):
    """ A COMBINE/OMEX archive

    Attributes:
        contents (:obj:`list` of :obj:`CombineArchiveContent`): contents of the archive
        description (:obj:`str`): description
        authors (:obj:`list` of :obj:`Person`): authors
        created (:obj:`datetime.datetime`): created date
        updated (:obj:`datetime.datetime`): updated date
    """

    def __init__(self, contents=None, description=None, authors=None, created=None, updated=None):
        """
        Args:
            contents (:obj:`list` of :obj:`CombineArchiveContent`, optional): contents of the archive
            description (:obj:`str`, optional): description
            authors (:obj:`list` of :obj:`Person`, optional): authors
            created (:obj:`datetime.datetime`, optional): created date
            updated (:obj:`datetime.datetime`, optional): updated date
        """
        self.contents = contents or []
        self.description = description
        self.authors = authors or []
        self.created = created
        self.updated = updated

    def get_master_content(self):
        """ Get the master content of an archive

        Returns:
            :obj:`list` of :obj:`CombineArchiveContent`: master content
        """
        master_content = []
        for content in self.contents:
            if content.master:
                master_content.append(content)
        return master_content

    def to_tuple(self):
        """ Tuple representation of a COMBINE/OMEX archive

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation of a COMBINE/OMEX archive
        """
        contents = tuple(none_sorted(content.to_tuple() for content in self.contents))
        authors = tuple(none_sorted(author.to_tuple() for author in self.authors))
        return (contents, self.description, authors, self.created, self.updated)

    def is_equal(self, other):
        """ Determine if two content items are equal

        Args:
            other (:obj:`CombineArchiveContent`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two archives are equal
        """
        return self.__class__ == other.__class__ \
            and are_lists_equal(self.contents, other.contents) \
            and self.description == other.description \
            and are_lists_equal(self.authors, other.authors) \
            and self.created == other.created \
            and self.updated == other.updated


class CombineArchiveContent(CombineArchiveBase):
    """ A content item (e.g., file) in a COMBINE/OMEX archive

    Attributes:
        location (:obj:`str`): path to the content
        format (:obj:`str`): URL for the specification of the format of the content
        master (:obj:`bool`): :obj:`True`, if the content is the "primary" content of the parent archive
        description (:obj:`str`): description
        authors (:obj:`list` of :obj:`Person`): authors
        created (:obj:`datetime.datetime`): created date
        updated (:obj:`datetime.datetime`): updated date
    """

    def __init__(self, location=None, format=None, master=False, description=None, authors=None, created=None, updated=None):
        """
        Args:
            location (:obj:`str`, optional): path to the content
            format (:obj:`str`, optional): URL for the specification of the format of the content
            master (:obj:`bool`, optional): :obj:`True`, if the content is the "primary" content of the parent archive
            description (:obj:`str`, optional): description
            authors (:obj:`list` of :obj:`Person`, optional): authors
            created (:obj:`datetime.datetime`, optional): created date
            updated (:obj:`datetime.datetime`, optional): updated date
        """
        self.location = location
        self.format = format
        self.master = master
        self.description = description
        self.authors = authors or []
        self.created = created
        self.updated = updated

    def to_tuple(self):
        """ Tuple representation of a content item of a COMBINE/OMEX archive

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation of a content item of a COMBINE/OMEX archive
        """
        authors = tuple(none_sorted(author.to_tuple() for author in self.authors))
        return (self.location, self.format, self.master, self.description, authors, self.created, self.updated)

    def is_equal(self, other):
        """ Determine if two content items are equal

        Args:
            other (:obj:`CombineArchiveContent`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two content items are equal
        """
        return self.__class__ == other.__class__ \
            and self.location == other.location \
            and self.format == other.format \
            and self.master == other.master \
            and self.description == other.description \
            and are_lists_equal(self.authors, other.authors) \
            and self.created == other.created \
            and self.updated == other.updated


class CombineArchiveContentFormat(str, enum.Enum):
    """ Format for the content of COMBINE/OMEX archives """
    BMP = 'http://purl.org/NET/mediatypes/image/bmp'
    BNGL = 'http://purl.org/NET/mediatypes/text/bngl+plain'
    BioPAX = 'http://identifiers.org/combine.specifications/biopax'
    CellML = 'http://identifiers.org/combine.specifications/cellml'
    CopasiML = 'http://purl.org/NET/mediatypes/application/x-copasi'
    CSV = 'http://purl.org/NET/mediatypes/text/csv'
    Escher = 'http://purl.org/NET/mediatypes/application/escher+json'
    GIF = 'http://purl.org/NET/mediatypes/image/gif'
    GINML = 'http://purl.org/NET/mediatypes/application/ginml+xml'
    HDF5 = 'http://purl.org/NET/mediatypes/application/x-hdf'
    HOC = 'http://purl.org/NET/mediatypes/text/x-hoc'
    IPython_Notebook = 'http://purl.org/NET/mediatypes/application/x-ipynb+json'
    JPEG = 'http://purl.org/NET/mediatypes/image/jpeg'
    JSON = 'http://purl.org/NET/mediatypes/application/json'
    Kappa = 'http://purl.org/NET/mediatypes/text/x-kappa'
    LEMS = 'http://purl.org/NET/mediatypes/application/lems+xml'
    MASS = 'http://purl.org/NET/mediatypes/application/mass+json'
    MATLAB = 'http://purl.org/NET/mediatypes/text/x-matlab'
    MATLAB_DATA = 'http://purl.org/NET/mediatypes/application/x-matlab-data'
    MorpheusML = 'http://purl.org/NET/mediatypes/application/morpheusml+xml'
    NeuroML = 'http://identifiers.org/combine.specifications/neuroml'
    NuML = 'http://purl.org/NET/mediatypes/application/numl+xml'
    OMEX = 'http://identifiers.org/combine.specifications/omex'
    OMEX_MANIFEST = 'http://identifiers.org/combine.specifications/omex-manifest'
    OMEX_METADATA = 'http://identifiers.org/combine.specifications/omex-metadata'
    OWL = 'http://purl.org/NET/mediatypes/application/rdf+xml'
    PDF = 'http://purl.org/NET/mediatypes/application/PDF'
    pharmML = 'http://purl.org/NET/mediatypes/application/pharmml+xml'
    PNG = 'http://purl.org/NET/mediatypes/image/png'
    Python = 'http://purl.org/NET/mediatypes/application/x-python-code'
    R = 'http://purl.org/NET/mediatypes/text/x-r'
    R_Project = 'http://purl.org/NET/mediatypes/application/x-r-project'
    RBA = 'http://purl.org/NET/mediatypes/application/rba+zip'
    SBGN = 'http://identifiers.org/combine.specifications/sbgn'
    SBML = 'http://identifiers.org/combine.specifications/sbml'
    SBOL = 'http://identifiers.org/combine.specifications/sbol'
    SBOL_VISUAL = 'http://identifiers.org/combine.specifications/sbol-visual'
    Scilab = 'http://purl.org/NET/mediatypes/application/x-scilab'
    SED_ML = 'http://identifiers.org/combine.specifications/sed-ml'
    SimBiology_Project = 'http://purl.org/NET/mediatypes/application/x-sbproj'
    Smoldyn = 'http://purl.org/NET/mediatypes/text/smoldyn+plain'
    SVG = 'http://purl.org/NET/mediatypes/image/svg+xml'
    TEXT = 'http://purl.org/NET/mediatypes/text/plain'
    TIFF = 'http://purl.org/NET/mediatypes/image/tiff'
    TSV = 'http://purl.org/NET/mediatypes/text/tab-separated-values'
    VCML = 'http://purl.org/NET/mediatypes/application/vcml+xml'
    Vega = 'http://purl.org/NET/mediatypes/application/vnd.vega.v5+json'
    Vega_Lite = 'http://purl.org/NET/mediatypes/application/vnd.vegalite.v3+json'
    WEBP = 'http://purl.org/NET/mediatypes/image/webp'
    XML = 'http://purl.org/NET/mediatypes/application/xml'
    XPP = 'http://purl.org/NET/mediatypes/text/x-xpp'
    XLSX = 'http://purl.org/NET/mediatypes/application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    YAML = 'http://purl.org/NET/mediatypes/application/x-yaml'
    ZGINML = 'http://purl.org/NET/mediatypes/application/zginml+zip'
    ZIP = 'http://purl.org/NET/mediatypes/application/zip'
    OTHER = 'http://purl.org/NET/mediatypes/application/octet-stream'


class CombineArchiveContentFormatPattern(str, enum.Enum):
    """ Format for the content of COMBINE/OMEX archives """
    BMP = r'^https?://purl\.org/NET/mediatypes/image/bmp$'
    BNGL = r'^https?://purl\.org/NET/mediatypes/text/bngl\+plain($|\.)'
    BioPAX = r'^https?://identifiers\.org/combine\.specifications/biopax($|\.)'
    CellML = r'^https?://identifiers\.org/combine\.specifications/cellml($|\.)'
    CopasiML = r'^https?://purl\.org/NET/mediatypes/application/x-copasi$'
    CSV = r'^https?://purl\.org/NET/mediatypes/text/csv$'
    Escher = r'^https?://purl\.org/NET/mediatypes/application/escher\+json$'
    GIF = r'^https?://purl\.org/NET/mediatypes/image/gif$'
    GINML = r'^https?://purl\.org/NET/mediatypes/application/ginml\+xml$'
    HDF5 = r'^https?://purl\.org/NET/mediatypes/application/x-hdf5?$'
    HOC = r'^https?://purl\.org/NET/mediatypes/text/x-hoc$'
    IPython_Notebook = r'^https?://purl\.org/NET/mediatypes/application/x-ipynb\+json$'
    JPEG = r'^https?://purl\.org/NET/mediatypes/image/jpeg$'
    JSON = r'^https?://purl\.org/NET/mediatypes/application/json$'
    Kappa = r'^https?://purl\.org/NET/mediatypes/text/x-kappa$'
    LEMS = r'^https?://purl\.org/NET/mediatypes/application/lems\+xml$'
    MASS = r'^https?://purl\.org/NET/mediatypes/application/mass\+json$'
    MATLAB = r'^https?://purl\.org/NET/mediatypes/text/x-matlab$'
    MATLAB_DATA = r'^https?://purl\.org/NET/mediatypes/application/x-matlab-data$'
    MorpheusML = r'^https?://purl\.org/NET/mediatypes/application/morpheusml\+xml$'
    NeuroML = r'^https?://identifiers\.org/combine\.specifications/neuroml($|\.)'
    NuML = r'^https?://purl\.org/NET/mediatypes/application/numl\+xml$'
    OMEX = r'^https?://identifiers\.org/combine\.specifications/omex($|\.)'
    OMEX_MANIFEST = r'^https?://identifiers\.org/combine\.specifications/omex-manifest($|\.)'
    OMEX_METADATA = r'^https?://identifiers\.org/combine\.specifications/omex-metadata($|\.)'
    OWL = r'^https?://purl\.org/NET/mediatypes/application/rdf\+xml$'
    PDF = r'^https?://purl\.org/NET/mediatypes/application/pdf$'
    pharmML = r'^https?://purl\.org/NET/mediatypes/application/pharmml\+xml$'
    PNG = r'^https?://purl\.org/NET/mediatypes/image/png$'
    Python = r'^https?://purl\.org/NET/mediatypes/application/x-python-code$'
    R = r'^https?://purl\.org/NET/mediatypes/text/x-r$'
    R_Project = r'^https?://purl\.org/NET/mediatypes/application/x-r-project$'
    RBA = r'^https?://purl\.org/NET/mediatypes/application/rba+zip$'
    SBGN = r'^https?://identifiers\.org/combine\.specifications/sbgn($|\.)'
    SBML = r'^https?://identifiers\.org/combine\.specifications/sbml($|\.)'
    SBOL = r'^https?://identifiers\.org/combine\.specifications/sbol($|\.)'
    SBOL_VISUAL = r'^https?://identifiers\.org/combine\.specifications/sbol-visual($|\.)'
    Scilab = r'^https?://purl\.org/NET/mediatypes/application/x-scilab$'
    SED_ML = r'^https?://identifiers\.org/combine\.specifications/sed\-?ml($|\.)'
    SimBiology_Project = r'^https?://purl\.org/NET/mediatypes/application/x-sbproj$'
    Smoldyn = r'^https?://purl\.org/NET/mediatypes/text/smoldyn\+plain$'
    SVG = r'^https?://purl\.org/NET/mediatypes/image/svg\+xml$'
    TEXT = r'^https?://purl\.org/NET/mediatypes/text/plain$'
    TIFF = r'^https?://purl\.org/NET/mediatypes/image/tiff$'
    TSV = r'^https?://purl\.org/NET/mediatypes/text/tab-separated-values$'
    Vega = r'^https?://purl\.org/NET/mediatypes/application/vnd\.vega\.v5\+json$'
    Vega_Lite = r'^https?://purl\.org/NET/mediatypes/application/vnd\.vegalite\.v3\+json$'
    VCML = r'^https?://purl\.org/NET/mediatypes/application/vcml\+xml$'
    WEBP = r'^https?://purl\.org/NET/mediatypes/image/webp$'
    XLSX = r'^https?://purl\.org/NET/mediatypes/application/vnd\.openxmlformats-officedocument\.spreadsheetml\.sheet$'
    XML = r'^https?://purl\.org/NET/mediatypes/application/xml$'
    XPP = r'^https?://purl\.org/NET/mediatypes/text/x-xpp$'
    YAML = r'^https?://purl\.org/NET/mediatypes/application/x-yaml$'
    ZGINML = r'^https?://purl\.org/NET/mediatypes/application/zginml\+zip$'
    ZIP = r'^https?://purl\.org/NET/mediatypes/application/zip$'
    OTHER = r'^https?://purl\.org/NET/mediatypes/application/octet-stream$'



""" Utilities for validating BNGL models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-04-13
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ...log.data_model import StandardOutputErrorCapturerLevel
from ...log.utils import StandardOutputErrorCapturer
import bionetgen
import os
import shutil
import tempfile


def validate_model(filename, name=None):
    """ Check that a model is valid

    Args:
        filename (:obj:`str`): path to model
        name (:obj:`str`, optional): name of model for use in error messages

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`bionetgen.xmlapi.model.bngmodel`: model
    """
    errors = []
    warnings = []
    model = None

    if filename:
        if os.path.isfile(filename):
            # create temporary file with ``.bngl`` extension because pyBioNet requires this
            if os.path.splitext(filename)[1] == '.bngl':
                bngl_filename = filename
            else:
                fid, bngl_filename = tempfile.mkstemp(dir=os.path.dirname(filename), suffix='.bngl')
                os.close(fid)
                shutil.copyfile(filename, bngl_filename)

            # parse model
            model, temp_errors, stdout = read_model(bngl_filename, filename)
            errors.extend(temp_errors)

            if model and not model.active_blocks:
                msg = '`{}` does not appear to a valid BNGL or BNGL XML file because it does not have any BNGL blocks.'.format(filename)
                errors.append([msg])

            # clean up temporary file
            if bngl_filename != filename:
                os.remove(bngl_filename)

        else:
            errors.append(['`{}` is not a file.'.format(filename or '')])

    else:
        errors.append(['`filename` must be a path to a file, not `{}`.'.format(filename or '')])

    return (errors, warnings, model)


def read_model(bngl_filename, filename):
    model = None
    errors = []
    with StandardOutputErrorCapturer(level=StandardOutputErrorCapturerLevel.c, relay=False) as captured:
        try:
            model = bionetgen.bngmodel(bngl_filename)
        except Exception as exception:
            errors.append(['`{}` is not a valid BNGL or BGNL XML file.'.format(filename or ''), [[str(exception)]]])
    return model, errors, captured.get_text()



""" Miscellaneous utilities

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..data_model import ValueType, OntologyTerm
from ..warnings import warn, BioSimulatorsWarning
import json
import numpy
import re

__all__ = [
    'are_lists_equal', 'none_sorted', 'assert_exception',
    'validate_value', 'validate_str_value', 'format_value', 'parse_value',
    'patch_dict', 'pad_arrays_to_consistent_shapes',
    'flatten_nested_list_of_strings',
    'raise_errors_warnings',
    'format_float',
]


def are_lists_equal(a, b):
    """ Determine if two lists are equal, optionally up to the order of the elements

    Args:
        a (:obj:`list`): first list
        b (:obj:`list`): second list

    Returns:
        :obj:`bool`: :obj:`True`, if lists are equal
    """
    if len(a) != len(b):
        return False

    a = none_sorted(a, key=lambda x: x.to_tuple())
    b = none_sorted(b, key=lambda x: x.to_tuple())

    for a_el, b_el in zip(a, b):
        if not a_el.is_equal(b_el):
            return False

    return True


def none_sort_key_gen(key=None):
    class NoneComparator(object):
        def __init__(self, obj):
            if key:
                self.obj = key(obj)
            else:
                self.obj = obj

        def __lt__(self, other):
            return none_comparator(self.obj, other.obj) < 0

        def __gt__(self, other):
            return none_comparator(self.obj, other.obj) > 0

        def __eq__(self, other):
            return none_comparator(self.obj, other.obj) == 0

        def __le__(self, other):
            return none_comparator(self.obj, other.obj) <= 0

        def __ge__(self, other):
            return none_comparator(self.obj, other.obj) >= 0

        def __ne__(self, other):
            return none_comparator(self.obj, other.obj) != 0
    return NoneComparator


def none_comparator(x, y):
    if x == y:
        return 0

    if isinstance(x, tuple) and not isinstance(y, tuple):
        return 1
    if not isinstance(x, tuple) and isinstance(y, tuple):
        return -1
    if isinstance(x, tuple) and isinstance(y, tuple):
        if len(x) < len(y):
            return -1
        if len(x) > len(y):
            return 1

        for x1, y1 in zip(x, y):
            cmp = none_comparator(x1, y1)
            if cmp != 0:
                return cmp

        return 0  # pragma: no cover

    if x is None:
        return -1
    if y is None:
        return 1

    if x < y:
        return -1
    if x > y:
        return 1


def none_sorted(arr, key=None):
    """ Sort an error that contains :obj:`None`

    Args:
        arr (:obj:`list`): array

    Returns:
        :obj:`list`: sorted array
    """
    return sorted(arr, key=none_sort_key_gen(key))


def assert_exception(success, exception):
    """ Raise an error if :obj:`success` is :obj:`False`

    Args:
        success (:obj:`bool`)
        exception (:obj:`Exception`)

    Raises:
        :obj:`Exception`
    """
    if not success:
        raise exception


def validate_value(val, type):
    """ Determine if a value is a valid instance of type :obj:`type`

    Args:
        val (:obj:`object`): value
        type (:obj:`ValueType`): type

    Returns:
        :obj:`bool`: :obj:`True`, if a value is a valid instance of type :obj:`type`
    """
    if type == ValueType.boolean:
        return isinstance(val, bool)
    if type == ValueType.integer:
        return isinstance(val, int) and not isinstance(val, bool)
    if type == ValueType.float:
        return isinstance(val, float)
    if type == ValueType.string:
        return isinstance(val, str)
    if type == ValueType.kisao_id:
        return isinstance(val, OntologyTerm) and val.namespace == 'KISAO' and re.match(r'^KISAO_\d{7}$', val.id) is not None
    if type == ValueType.list:
        return isinstance(val, list)
    if type == ValueType.object:
        return isinstance(val, dict)
    if type == ValueType.any:
        return True
    raise NotImplementedError('Type {} is not supported'.format(type))


def validate_str_value(str_val, type):
    """ Determine if a value is a valid string representation of type :obj:`type`

    Args:
        str_val (:obj:`str`): string representation of a value
        type (:obj:`ValueType`): type

    Returns:
        :obj:`bool`: :obj:`True`, if a value is a valid string representation of type :obj:`type`
    """
    if type == ValueType.boolean:
        return str_val.lower() in ['true', 'false', '0', '1']
    if type == ValueType.integer:
        try:
            int(str_val)
            return True
        except ValueError:
            return False
    if type == ValueType.float:
        try:
            float(str_val)
            return True
        except ValueError:
            return False
    if type == ValueType.string:
        return True
    if type == ValueType.kisao_id:
        return re.match(r'^KISAO_\d{7}$', str_val)
    if type == ValueType.list:
        try:
            return isinstance(json.loads(str_val), list)
        except json.JSONDecodeError:
            return False
    if type == ValueType.object:
        try:
            return isinstance(json.loads(str_val), dict)
        except json.JSONDecodeError:
            return False
    if type == ValueType.any:
        try:
            json.loads(str_val)
            return True
        except json.JSONDecodeError:
            return False
    raise NotImplementedError('Type {} is not supported'.format(type))


def format_value(val, type):
    """ Format a value as a string (e.g., for use with an attribute of an XML object)

    Args:
        val (:obj:`object`): value
        type (:obj:`ValueType`): type

    Returns:
        :obj:`str`: string representation of the value
    """
    if type == ValueType.boolean:
        return str(val).lower()
    if type == ValueType.integer or type == ValueType.float:
        return str(val)
    if type == ValueType.string:
        return val
    if type == ValueType.kisao_id:
        return val.id
    if type == ValueType.list or type == ValueType.object or type == ValueType.any:
        return json.dumps(val)
    raise NotImplementedError('Type {} is not supported'.format(type))


def parse_value(str_val, type):
    """ Parse the string representation of a value (e.g., for use with an attribute of an XML object)

    Args:
        str_val (:obj:`str`): string representation of the value
        type (:obj:`ValueType`): type

    Returns:
        :obj:`object`: Python representation of the value
    """
    if type == ValueType.boolean:
        return str_val.lower() == 'true' or str_val.lower() == '1'
    if type == ValueType.integer:
        return int(str_val)
    if type == ValueType.float:
        return float(str_val)
    if type == ValueType.string:
        return str_val
    if type == ValueType.kisao_id:
        return OntologyTerm(
            namespace='KISAO',
            id=str_val,
            url='https://www.ebi.ac.uk/ols/ontologies/kisao/terms?iri=http%3A%2F%2Fwww.biomodels.net%2Fkisao%2FKISAO%23' + str_val,
        )
    if type == ValueType.list or type == ValueType.object or type == ValueType.any:
        return json.loads(str_val)
    raise NotImplementedError('Type {} is not supported'.format(type))


def patch_dict(dictionary, patch):
    """ Recursively patch the attributes of a dictionary

    Args:
        dictionary (:obj:`dict`): dictionary
        patch (:obj:`dict`): patch
    """
    patch_queue = [(dictionary, patch)]

    while patch_queue:
        props, props_patch = patch_queue.pop()
        for key, new_val in props_patch.items():
            if isinstance(props, dict):
                value = props.get(key, None)
            else:
                if key >= len(props):
                    props.extend([None] * (key + 1 - len(props)))
                    value = None
                else:
                    value = props[key]

            if isinstance(value, dict):
                patch_queue.append((props[key], new_val))

            elif isinstance(value, list) and isinstance(new_val, dict):
                patch_queue.append((props[key], new_val))

            else:
                props[key] = new_val


def pad_arrays_to_consistent_shapes(arrays):
    """ Pad a list of NumPy arrays to a consistent shape

    Args:
        arrays (:obj:`list` of :obj:`numpy.ndarray`): list of NumPy arrays

    Returns:
        :obj:`list` of :obj:`numpy.ndarray`: list of padded arrays
    """
    shapes = set()
    for array in arrays:
        if array is not None:
            shape = array.shape
            if not shape and array.size:
                shape = (1,)
            shapes.add(shape)

    if len(shapes) > 1:
        warn('Arrays do not have consistent shapes', UserWarning)

    max_shape = []
    for shape in shapes:
        max_shape = max_shape + [1 if max_shape else 0] * (len(shape) - len(max_shape))
        shape = list(shape) + [1 if shape else 0] * (len(max_shape) - len(shape))
        max_shape = [max(x, y) for x, y in zip(max_shape, shape)]

    padded_arrays = []
    for array in arrays:
        if array is None:
            array = numpy.full(max_shape, numpy.nan)

        shape = tuple(list(array.shape)
                      + [1 if array.size else 0]
                      * (len(max_shape) - array.ndim))
        array = array.astype('float64').reshape(shape)

        pad_width = tuple((0, x - y) for x, y in zip(max_shape, shape))

        if pad_width:
            array = numpy.pad(array,
                              pad_width,
                              mode='constant',
                              constant_values=numpy.nan)

        padded_arrays.append(array)

    return padded_arrays


def flatten_nested_list_of_strings(nested_list, prefix='- ', indent=' ' * 2):
    """ Flatten a nested list of strings

    Args:
        nested_list (nested :obj:`list` of :obj:`str`): nested list of string
        prefix (:obj:`str`, optional): prefix
        indentation (:obj:`str`, optional): indentation

    Returns:
        :obj:`str`: flattened string
    """
    flattened = []
    for item in nested_list:
        flattened.append(prefix + item[0].replace('\n', '\n' + ' ' * len(prefix)))
        if len(item) > 1:
            flattened.append(
                indent
                + flatten_nested_list_of_strings(item[1], prefix=prefix, indent=indent).replace('\n', '\n' + indent)
            )

    return '\n'.join(flattened)


def raise_errors_warnings(errors, warnings=None, error_summary=None, warning_summary=None):
    """ Raises errors and warnings

    Args:
        errors (nested :obj:`list` of :obj:`str`): errors
        warnings (nested :obj:`list` of :obj:`str`): warnings
        error_summary (:obj:`str`, optional): summary of errors
        warning_summary (:obj:`str`, optional): summary of warnings

    Raises:
        :obj:`ValueError`: if errors
    """
    if warnings:
        msg = flatten_nested_list_of_strings(warnings)
        if warning_summary:
            msg = warning_summary + '\n  ' + msg.replace('\n', '\n  ')
        warn(msg, BioSimulatorsWarning)

    if errors:
        msg = flatten_nested_list_of_strings(errors)
        if error_summary:
            msg = error_summary + '\n  ' + msg.replace('\n', '\n  ')
        raise ValueError(msg)


def format_float(val):
    """ Format a float in scientific notation

    Args:
        val (:obj:`float`): value

    Returns:
        :obj:`str`: value formatted as a string
    """
    if val == int(val):
        return str(int(val))

    elif abs(val) < 1e3 and abs(val) > 1e-3:
        return str(val)

    else:
        return (
            '{:e}'
            .format(val)
            .replace('e-0', 'e-')
            .replace('e+0', 'e')
            .replace('e+', 'e')
        )



""" Utilities for working with Vega

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-07-22
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

__all__ = ['dict_to_vega_dataset']


def dict_to_vega_dataset(data_dict):
    """ Convert a dictionary of values, such as that produced by COBRApy, into a Vega
    dataset (array of data elements)

    Args:
        data_dict (:obj:`dict`): dictionary that maps the labels of predicted variables
            to their predicted values

    Returns:
        :obj:`list`: list of data elements, each with two keys ``label`` and `values`` whose values
            are the labels of each variable and an array of their predicted values
    """
    data_set = []
    for id, val in data_dict.items():
        data_set.append({
            'label': id,
            'values': val if isinstance(val, list) else [val]
        })
    return data_set



""" Methods for validating XPP files

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-08-06
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import collections
import glob
import os
import re
import subprocess
import tempfile


__all__ = ['validate_model', 'get_xpp_input_configuration_from_directory']


def validate_model(filename,
                   set_filename=None, parameter_filename=None, initial_conditions_filename=None,
                   name=None):
    """ Check that a model is valid

    Args:
        filename (:obj:`str`): path to model file or directory with ``.ode`` and possibly set (``.set``),
            parameter (``.par``), andd initial conditions (``.ic``) files
        set_filename (:obj:`str`, optional): path to XPP set file
        parameter_filename (:obj:`str`, optional): path to XPP parameters file
        initial_conditions_filename (:obj:`str`, optional): path to XPP initial conditions file
        name (:obj:`str`, optional): name of model for use in error messages

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`dict`: values of parameters of initial conditions of variables
    """
    errors = []
    warnings = []
    simulation = None

    if filename and os.path.isdir(filename):
        (
            filename, temp_set_filename, temp_parameter_filename,
            temp_initial_conditions_filename,
        ) = get_xpp_input_configuration_from_directory(filename)
        set_filename = set_filename or temp_set_filename
        parameter_filename = parameter_filename or temp_parameter_filename
        initial_conditions_filename = initial_conditions_filename or temp_initial_conditions_filename

    # check model file exists and is valid
    if not isinstance(filename, str):
        errors.append(['`{}` is not a path to an XPP file.'.format(filename)])
        return (errors, warnings, simulation)

    if not os.path.isfile(filename):
        errors.append(['XPP file `{}` does not exist.'.format(filename)])
        return (errors, warnings, simulation)

    var_param_fid, var_param_filename = tempfile.mkstemp()
    os.close(var_param_fid)

    cmd = ['xppaut', filename, '-qics', '-qpars', '-outfile', var_param_filename, '-quiet', '0']
    if set_filename is not None:
        cmd.append('-setfile')
        cmd.append(set_filename)
    if parameter_filename is not None:
        cmd.append('-parfile')
        cmd.append(parameter_filename)
    if initial_conditions_filename is not None:
        cmd.append('-icfile')
        cmd.append(initial_conditions_filename)
    result = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=False,
    )

    stdout = result.stdout.decode(errors='ignore').strip()

    if result.returncode != 0:
        errors.append(['`{}` could not be validated.'.format(filename), [[stdout]]])

    elif re.search(r'\berror\b', stdout, re.IGNORECASE):
        errors.append(['`{}` is not a valid XPP file.'.format(filename), [[stdout]]])

    else:
        cmd = ['xppaut', filename, '-qics', '-qpars', '-outfile', var_param_filename, '-quiet', '1']
        result = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=False,
        )
        stdout = result.stdout.decode(errors='ignore').strip()
        if stdout:
            warnings.append([
                'The XPP file may be not be formulated correctly',
                [
                    [line] for line in stdout.split('\n')
                ],
            ])

        simulation = {
            'parameters': collections.OrderedDict(),
            'initial_conditions': collections.OrderedDict(),
            'sets': {},
            'auxiliary_variables': collections.OrderedDict(),
            'simulation_method': {},
            'plot': {},
        }
        block = None
        with open(var_param_filename, 'r') as file:
            for line in file:
                line = line.strip()
                if line.startswith('#'):
                    if line == '#Parameters query:':
                        block = 'parameters'
                    if line == '#Initial conditions query:':
                        block = 'initial_conditions'
                elif block:
                    id, _, value = line.partition(' ')
                    simulation[block][id] = float(value)

        el_keys = ['xp', 'yp', 'zp']
        for i_el in range(7):
            el_keys.append('xp' + str(i_el + 2))
            el_keys.append('yp' + str(i_el + 2))
            el_keys.append('zp' + str(i_el + 2))

        with open(filename, 'r') as file:
            for line in file:
                if line.startswith('aux '):
                    name, _, expr = line[4:].strip().partition('=')
                    simulation['auxiliary_variables'][name] = expr

                if line.startswith('set '):
                    name, _, values = line[4:].strip().partition(' ')
                    simulation['sets'][name.strip().lower()] = {
                        'parameters': {},
                        'initial_conditions': {},
                    }
                    for val in values[1:-1].split(','):
                        paramVar, _, val = val.partition('=')
                        paramVar = paramVar.strip()
                        val = float(val.strip())

                        if paramVar in simulation['parameters']:
                            paramVar = paramVar.lower()
                            simulation['sets'][name.strip().lower()]['parameters'][paramVar] = val
                            # simulation['parameters'][paramVar] = val
                        else:
                            paramVar = paramVar.upper()
                            simulation['sets'][name.strip().lower()]['initial_conditions'][paramVar] = val
                            simulation['initial_conditions'][paramVar] = val

                elif line.startswith('@'):
                    line = line[1:]
                    i_comment = line.find('#')
                    if i_comment > -1:
                        line = line[0:i_comment]
                    for cmd in line.split(','):
                        parts = cmd.split('=')
                        if len(parts) > 1:
                            key = parts[0].lstrip()
                            val = parts[1].rstrip()
                            if ' ' not in key and ' ' not in val:
                                key = key.lower()
                                if key == 'nout':
                                    key = 'njmp'

                                if key in [
                                    'xlo', 'xhi', 'ylo', 'yhi',
                                    'xmin', 'xmax', 'ymin', 'ymax', 'zmin', 'zmax',
                                    'phi', 'theta',
                                    'nplot',
                                ]:
                                    simulation['plot'][key] = float(val)

                                elif key in [
                                    'lt', 'axes', 'nplot',
                                ]:
                                    simulation['plot'][key] = int(float(val))

                                elif key in el_keys:
                                    axis = key[0]
                                    i_el = int(float(key[2:] or '1'))

                                    if 'elements' not in simulation['plot']:
                                        simulation['plot']['elements'] = {}
                                    if i_el not in simulation['plot']['elements']:
                                        simulation['plot']['elements'][i_el] = {}
                                    simulation['plot']['elements'][i_el][axis] = val.upper()

                                else:
                                    simulation['simulation_method'][key] = val

                elif line.startswith('d') and not ('=' in line and (' ' not in line or line.find('=') < line.find(' '))):
                    # check for "done" line; note just the singular character ``d`` defines the "done" line
                    break

        if set_filename is not None:
            with open(set_filename, 'r') as file:
                block = None
                for line in file:
                    if line.startswith('#'):
                        block = line[1:].strip()
                    elif line:
                        if block == 'Old ICs':
                            val, _, var = line.partition(' ')
                            val = float(val.strip())
                            var = var.strip().upper()
                            simulation['initial_conditions'][var] = val

                        elif block == 'Parameters':
                            val, _, var = line.partition(' ')
                            val = float(val.strip())
                            var = var.strip().lower()
                            simulation['parameters'][var] = val

        if parameter_filename is not None:
            with open(parameter_filename, 'r') as file:
                file.readline()
                for line in file:
                    val, _, param = line.partition(' ')
                    simulation['parameters'][param.strip()] = float(val.strip())

        if initial_conditions_filename is not None:
            with open(initial_conditions_filename, 'r') as file:
                for val, var in zip(file, simulation['initial_conditions'].keys()):
                    simulation['initial_conditions'][var] = float(val.strip())

        t_0 = simulation['simulation_method'].get('t0', 0.)
        try:
            t_0 = float(t_0)
        except ValueError:
            errors.append(['T0 must be a float, not `{}`'.format(t_0)])

        duration = simulation['simulation_method'].get('total', 20.)
        try:
            duration = float(duration)
        except ValueError:
            errors.append(['TOTAL must be a float, not `{}`'.format(duration)])

        d_t = simulation['simulation_method'].get('dt', 0.05)
        try:
            d_t = float(d_t)
        except ValueError:
            errors.append(['DT must be a float, not `{}`'.format(d_t)])

        n_jmp = simulation['simulation_method'].get('njmp', 1)
        try:
            n_jmp = float(n_jmp)
            if n_jmp != int(n_jmp) or n_jmp < 1:
                errors.append(['NJMP must be a positive integer, not `{}`'.format(n_jmp)])
        except ValueError:
            errors.append(['NJMP must be a positive integer, not `{}`'.format(n_jmp)])

        if not errors:
            number_of_steps = duration / (d_t * n_jmp)
            if (number_of_steps % 1.) > 1e-8:
                errors.append([
                    'Number of steps must be an integer',
                    [
                        ['t0: {}'.format(t_0)],
                        ['total: {}'.format(duration)],
                        ['dt: {}'.format(d_t)],
                        ['njmp: {}'.format(n_jmp)],
                    ],
                ])

    os.remove(var_param_filename)

    return (errors, warnings, simulation)


def get_xpp_input_configuration_from_directory(dirname):
    """ Get input file configuration from a directory of files

    * ``*.ode``: Main model model
    * ``*.set``: Set file
    * ``*.par``: Parameters file
    * ``*.ic``: Initial conditions file

    Args:
        dirname (:obj:`str`): path to directory of XPP files

    Returns:
        :obj:`tuple`:

            * :obj:`str`: path to main model file
            * :obj:`str`: path to set file
            * :obj:`str`: path to parameters file
            * :obj:`str`: path to initial conditions file
            * :obj:`str`: name of a set to use
            * :obj:`str`: name of a set not to use
    """
    errors = []

    ode_filenames = glob.glob(os.path.join(dirname, "*.ode"))
    if len(ode_filenames) > 1:
        errors.append('Only 1 ODE file can be used at a time.')
    elif ode_filenames:
        ode_filename = ode_filenames[0]
    else:
        errors.append('The directory must contain an ODE file.')

    set_filenames = glob.glob(os.path.join(dirname, "*.set"))
    if len(set_filenames) > 1:
        errors.append('Only 1 set file can be used at a time.')
    elif set_filenames:
        set_filename = set_filenames[0]
    else:
        set_filename = None

    parameter_filenames = glob.glob(os.path.join(dirname, "*.par"))
    if len(parameter_filenames) > 1:
        errors.append('Only 1 parameter file can be used at a time.')
    elif parameter_filenames:
        parameter_filename = parameter_filenames[0]
    else:
        parameter_filename = None

    initial_conditions_filenames = glob.glob(os.path.join(dirname, "*.ic"))
    if len(initial_conditions_filenames) > 1:
        errors.append('Only 1 initial conditions file can be used at a time.')
    elif initial_conditions_filenames:
        initial_conditions_filename = initial_conditions_filenames[0]
    else:
        initial_conditions_filename = None

    if errors:
        raise ValueError('`{}` does not contain a valid set of XPP files:\n  - {}'.format(dirname, '\n  - '.join(errors)))

    return (ode_filename, set_filename, parameter_filename, initial_conditions_filename)



""" Methods for generating OMEX metdata files for models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-07-19
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import OmexMetadataOutputFormat
import os
import pyomexmeta

__all__ = ['build_omex_meta_file_for_model', 'get_local_combine_archive_content_uri', 'get_global_combine_archive_content_uri']


def build_omex_meta_file_for_model(model_filename,
                                   metadata_filename,
                                   metadata_format=OmexMetadataOutputFormat.rdfxml_abbrev):
    """ Create an OMEX metadata file for a model encoded in CellML or SBML. Also
    add missing metadata ids to the model file.

    Args:
        model_filename (:obj:`str`): path to model to extract metadata about
        metadata_filename (:obj:`str`): path to save metadata
        metadata_format (:obj:`OmexMetadataOutputFormat`, optional): format for :obj:`metadata_filename`
    """
    # uses subprocess by pyomexmetadata has no error handling
    if not isinstance(model_filename, str) or not os.path.isfile(model_filename):
        raise FileNotFoundError('`{}` is not a file.'.format(model_filename))

    if not isinstance(metadata_filename, str) or not os.path.isdir(os.path.dirname(metadata_filename)):
        raise FileNotFoundError('The parent directory for `{}` does not exist.'.format(metadata_filename))

    if not isinstance(metadata_format, OmexMetadataOutputFormat):
        raise NotImplementedError('Output format `{}` is not supported.'.format(metadata_format))

    rdf = pyomexmeta.RDF()

    pyomexmeta_log_level = pyomexmeta.Logger.get_level()
    pyomexmeta.Logger.clear()
    pyomexmeta.Logger.set_level(pyomexmeta.eLogLevel.err)

    editor = rdf.to_editor(model_filename, generate_new_metaids=True, sbml_semantic_extraction=True)

    pyomexmeta.Logger.set_level(pyomexmeta_log_level)

    logger = pyomexmeta.Logger()

    errors = ''.join([logger[i_message].get_message() for i_message in range(len(logger))])
    if errors:
        raise ValueError('Metadata could not be extracted from model `{}`:\n  {}'.format(
            model_filename, '\n'.join(errors).replace('\n', '\n  ')))

    model = editor.get_xml()
    if not model:
        errors = ''.join([logger[i_message].get_message() for i_message in range(len(logger))])
        raise RuntimeError('Model `{}` could not be read:\n  {}'.format(
            model_filename, errors.replace('\n', '\n  ')))
    with open(model_filename, 'w') as file:
        file.write(model)

    if rdf.to_file(metadata_filename, metadata_format.value) != 0:
        errors = ''.join([logger[i_message].get_message() for i_message in range(len(logger))])
        raise RuntimeError('OMEX metadata could not be saved to `{}` in `{}` format:\n  {}'.format(
            metadata_filename, metadata_format.value, errors.replace('\n', '\n  ')))


def get_local_combine_archive_content_uri(content_uri, archive_uri=None):
    """ Get the relative URI for a content item of a COMBINE/OMEX archive

    Args:
        content_uri (:obj:`str`): global URI for a content item of a COMBINE/OMEX archive
        archive_uri (:obj:`str`, optional): URI for the parent COMBINE/OMEX archive

    Returns:
        :obj:`tuple:`

            * :obj:`str`: global URI for the content item
            * :obj:`str`: URI for the parent COMBINE/OMEX archive
    """
    if archive_uri:
        if content_uri in [archive_uri, archive_uri + '/']:
            return ('.', archive_uri)
        if content_uri.startswith(archive_uri + '/'):
            return ('./' + os.path.relpath(content_uri[len(archive_uri)+1:], '.'), archive_uri)
        return (content_uri, None)
    else:
        return (content_uri, archive_uri)


def get_global_combine_archive_content_uri(content_rel_uri, archive_uri=None):
    """ Get a global URI for a content item of a COMBINE/OMEX archive

    Args:
        content_rel_uri (:obj:`str`): URI for a content item of a COMBINE/OMEX archive,
            relative to its parent archive
        archive_uri (:obj:`str`, optional): URI for the parent COMBINE/OMEX archive

    Returns:
        :obj:`str`: global URI for the content item
    """
    if archive_uri:
        content_rel_uri = os.path.relpath(content_rel_uri, '.')
        if content_rel_uri == '.':
            return archive_uri
        else:
            return archive_uri + '/' + content_rel_uri
    else:
        return content_rel_uri



""" Utilities for working with KiSAO

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-05-27
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from kisao import Kisao
from kisao.data_model import TermType  # noqa: F401
import kisao.utils
import pronto  # noqa: F401
import re

__all__ = [
    'get_term',
    'get_term_type',
]


def get_term(id):
    """ Get a KiSAO term

    Args:
        id (:obj:`str`): id

    Returns:
        :obj:`pronto.Term`
    """
    if not id or not re.match(r'^KISAO_\d{7}$', id):
        return None
    kisao = Kisao()
    try:
        return kisao.get_term(id)
    except ValueError:
        return None


def get_term_type(term):
    """ Get the type of a KiSAO term (algorithm, algorithm characteristic, algorithm parameter)

    Args:
        term (:obj:`pronto.Term`): term

    Returns:
        :obj:`TermType`: type of the term
    """
    if term:
        return kisao.utils.get_term_type(term)

    return None



""" Warnings for misuse of SED-ML

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..warnings import BioSimulatorsWarning

__all__ = [
    'IllogicalSedmlWarning',
    'InconsistentVariableShapesWarning',
    'NoTasksWarning',
    'NoOutputsWarning',
    'NoDataSetsWarning',
    'NoCurvesWarning',
    'NoSurfacesWarning',
    'SedmlFeatureNotSupportedWarning',
]


class IllogicalSedmlWarning(BioSimulatorsWarning):
    """ Warning that a SED document is illogical, such as when a report or plot contains
    no datasets, curves, or surfaces.
    """
    pass  # pragma: no cover


class InconsistentVariableShapesWarning(BioSimulatorsWarning):
    """ Warning that the variables of a data generator have different shapes. """
    pass  # pragma: no cover


class NoTasksWarning(IllogicalSedmlWarning):
    """ Warning that a SED document does not have any tasks """
    pass


class NoOutputsWarning(IllogicalSedmlWarning):
    """ Warning that a SED document does not have any outputs """
    pass


class NoDataSetsWarning(IllogicalSedmlWarning):
    """ Warning that a report does not contain any data sets """
    pass


class NoCurvesWarning(IllogicalSedmlWarning):
    """ Warning that a 2D plot does not contain any curves """
    pass  # pragma: no cover


class NoSurfacesWarning(IllogicalSedmlWarning):
    """ Warning that a 3D plot does not contain any surfaces """
    pass  # pragma: no cover


class SedmlFeatureNotSupportedWarning(BioSimulatorsWarning):
    """ Warning that a feature of a SED document was skipped because it requires a SED feature that is not
    yet supported. The warning is only used when skipping a portion of a document doesn't affect the semantic
    meaning of the unskipped portions of the document (e.g., skipping generating a plot doesn't adversely
    affect reports).
    """
    pass  # pragma: no cover



""" Utilities for validating SBML models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-04-13
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import libsbml
import os


def validate_model(filename, name=None, validate_consistency=True):
    """ Check that a model is valid

    Args:
        filename (:obj:`str`): path to model
        name (:obj:`str`, optional): name of model for use in error messages
        validate_consistency (:obj:`str`, optional): whether to check the consistency of the model

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`libsbml.SBMLDocument`: model
    """
    errors = []
    warnings = []
    doc = None

    if filename:
        if os.path.isfile(filename):
            doc = libsbml.readSBMLFromFile(filename)
            if validate_consistency:
                doc.checkConsistency()

            for i_error in range(doc.getNumErrors()):
                sbml_error = doc.getError(i_error)
                if sbml_error.isInfo() or sbml_error.isWarning():
                    warnings.append([sbml_error.getMessage()])
                else:
                    errors.append([sbml_error.getMessage()])

        else:
            errors.append(['`{}` is not a file.'.format(filename or '')])

    else:
        errors.append(['`filename` must be a path to a file, not `{}`.'.format(filename or '')])

    return (errors, warnings, doc)



""" Methods for validating SED objects

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-09
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..kisao.utils import get_term as get_kisao_term, get_term_type as get_kisao_term_type
from ..xml.utils import validate_xpaths_ref_to_unique_objects, eval_xpath
from .data_model import (SedIdGroupMixin, AbstractTask, Task, RepeatedTask,  # noqa: F401
                         Model, ModelLanguage, ModelLanguagePattern,
                         ModelChange, ComputeModelChange,
                         Simulation, OneStepSimulation, SteadyStateSimulation,
                         UniformTimeCourseSimulation, Algorithm, Variable,
                         Range, FunctionalRange, UniformRange, VectorRange,
                         SetValueComputeModelChange,
                         Report, Plot2D, Plot3D, DataGenerator,
                         Calculation)
from .math import compile_math, eval_math
from .utils import (append_all_nested_children_to_doc, get_range_len,
                    is_model_language_encoded_in_xml,
                    does_model_language_use_xpath_variable_targets,
                    get_models_referenced_by_task,
                    get_all_sed_objects,
                    get_data_generators_for_output, get_variables_for_data_generators,
                    get_model_changes_for_task,
                    get_task_results_shape)
from kisao.data_model import TermType as KisaoTermType
import collections
import copy
import lxml.etree
import math
import networkx
import os
import re

__all__ = [
    'validate_doc',
    'validate_reference',
    'validate_task',
    'validate_repeated_task_has_one_model',
    'validate_model',
    'validate_model_language',
    'validate_model_source',
    'validate_model_with_language',
    'validate_model_change_types',
    'validate_model_changes',
    'validate_simulation_type',
    'validate_simulation',
    'validate_algorithm',
    'validate_uniform_range',
    'validate_output',
    'validate_data_generator',
    'validate_data_generator_variables',
    'validate_target',
    'validate_target_xpaths',
    'validate_calculation',
]


def validate_doc(doc, working_dir, validate_semantics=True,
                 validate_models_with_languages=True,
                 validate_targets_with_model_sources=True):
    """ Validate a SED document

    Args:
        doc (:obj:`SedDocument`): SED document
        working_dir (:obj:`str`): working directory (e.g., for referencing model files)
        validate_semantics (:obj:`bool`, optional): if :obj:`True`, check that SED-ML is semantically valid
        validate_models_with_languages (:obj:`bool`, optional): if :obj:`True`, validate models
        validate_targets_with_model_sources (:obj:`bool`, optional): if :obj:`True`, validate targets against
            their models

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    # validate that required ids exist and are unique
    for child_type in ('models', 'simulations', 'data_generators', 'tasks', 'outputs'):
        children = getattr(doc, child_type)

        missing_ids = [[str(i_child + 1)] for i_child, child in enumerate(children) if not getattr(child, 'id', None)]
        if missing_ids:
            errors.append([
                '{} must have ids. The following {} are missing ids:'.format(
                    child_type[0].upper() + child_type[1:], child_type),
                missing_ids,
            ])

        if validate_semantics:
            id_counts = {}
            for child in children:
                child_id = getattr(child, 'id', None)
                if child_id:
                    if child_id in id_counts:
                        id_counts[child_id] += 1
                    else:
                        id_counts[child_id] = 1

            repeated_ids = sorted([child_id for child_id, count in id_counts.items() if count > 1])

            if repeated_ids:
                errors.append([
                    '{} must have unique ids. The following ids are repeated:'.format(child_type[0].upper() + child_type[1:]),
                    [[id] for id in repeated_ids],
                ])

    if validate_semantics:
        errors.extend(validate_unique_ids(doc))

        # validate the models, simulations, tasks, data generators are children of the SED document
        doc_copy = copy.deepcopy(doc)
        append_all_nested_children_to_doc(doc_copy)

        for child_type in ('models', 'simulations', 'tasks', 'data_generators', 'outputs'):
            not_children = set([child.id for child in getattr(doc_copy, child_type)]).difference(
                set([child.id for child in getattr(doc, child_type)]))
            if not_children:
                errors.append([
                    '{} must be direct children of the SED document. The following {} are not direct children of the document:'.format(
                        child_type[0].upper() + child_type[1:], child_type),
                    [[child] for child in sorted(not_children)],
                ])

        # model
        model_ids = [model.id for model in doc.models]
        for i_model, model in enumerate(doc.models):
            model_errors, model_warnings = validate_model(model, model_ids, working_dir,
                                                          validate_models_with_languages=validate_models_with_languages)

            # append errors/warnings to global lists of errors and warnings
            model_id = '`' + model.id + '`' if model and model.id else str(i_model + 1)

            if model_errors:
                errors.append(['Model {} is invalid.'.format(model_id), model_errors])

            if model_warnings:
                warnings.append(['Model {} may be invalid.'.format(model_id), model_warnings])

        # model sources are acyclic
        model_source_graph = networkx.DiGraph()

        for model in doc.models:
            if model.id:
                model_source_graph.add_node(model.id)

        for model in doc.models:
            if model.id and model.source and model.source.startswith('#'):
                model_source_graph.add_edge(model.id, model.source[1:])

        try:
            networkx.algorithms.cycles.find_cycle(model_source_graph)
            errors.append(['The sources of the models are defined cyclically. The sources of models must be acyclic.'])
        except networkx.NetworkXNoCycle:
            pass

        # validate the compute model changes are acyclic
        model_change_graph = networkx.DiGraph()

        for model in doc.models:
            if model.id:
                model_change_graph.add_node(model.id)

        for model in doc.models:
            for change in model.changes:
                if isinstance(change, ComputeModelChange):
                    for variable in change.variables:
                        if model.id and variable.model and variable.model != model and variable.model.id:
                            model_change_graph.add_edge(model.id, variable.model.id)

        try:
            networkx.algorithms.cycles.find_cycle(model_change_graph)
            errors.append(['The compute changes of the models are defined cyclically. The changes must be acyclic.'])
        except networkx.NetworkXNoCycle:
            pass

        # algorithms and parameters are described with valid ids of KiSAO terms
        for i_sim, sim in enumerate(doc.simulations):
            sim_errors, sim_warnings = validate_simulation(sim)
            if sim_errors:
                sim_id = '`' + sim.id + '`' if sim and sim.id else str(i_sim + 1)
                errors.append(['Simulation {} is invalid.'.format(sim_id), sim_errors])
            if sim_warnings:
                sim_id = '`' + sim.id + '`' if sim and sim.id else str(i_sim + 1)
                warnings.append(['Simulation {} may be invalid.'.format(sim_id), sim_warnings])

        # basic tasks reference a model and a simulation
        task_errors = {}
        task_warnings = {}
        for task in doc.tasks:
            task_errors[task] = {'ranges': [], 'other': []}
            task_warnings[task] = {'ranges': [], 'other': []}
            if isinstance(task, Task):
                task_errors[task]['other'].extend(validate_reference(task, 'Task', 'model', 'model'))
                task_errors[task]['other'].extend(validate_reference(task, 'Task', 'simulation', 'simulation'))
                task_errors[task]['other'].extend(validate_task(task))

        # sub tasks of repeated tasks reference tasks and have unique orders
        sub_task_graph = networkx.DiGraph()
        for task in doc.tasks:
            if task.id:
                sub_task_graph.add_node(task.id)

            if isinstance(task, RepeatedTask):
                if not task.sub_tasks:
                    msg = 'Repeated task must have at least one sub-task.'
                    task_errors[task]['other'].append([msg])

                for i_sub_task, sub_task in enumerate(task.sub_tasks):
                    if sub_task.task and isinstance(sub_task.task, AbstractTask):
                        if task.id and sub_task.task.id:
                            sub_task_graph.add_edge(task.id, sub_task.task.id)
                    else:
                        msg = 'Sub-task {} must reference a task.'.format(i_sub_task + 1)
                        task_errors[task]['other'].append([msg])

                duplicate_orders = [item for item, count in collections.Counter(
                    sub_task.order for sub_task in task.sub_tasks).items() if count > 1]
                if duplicate_orders:
                    msg = ('The `order` of each sub-task should be distinct within a repeated task. '
                           'Multiple sub-tasks have the following orders:\n  - {}').format(
                        '\n  - '.join([str(order) for order in sorted(duplicate_orders)]))
                    task_errors[task]['other'].append([msg])

        subtasks_cyclic = False
        try:
            networkx.algorithms.cycles.find_cycle(sub_task_graph)
            subtasks_cyclic = True
            errors.append(['The subtasks are defined cyclically. The graph of subtasks must be acyclic.'])
        except networkx.NetworkXNoCycle:
            for task in doc.tasks:
                if isinstance(task, RepeatedTask):
                    sub_task_shapes = set()
                    for sub_task in task.sub_tasks:
                        sub_task_shapes.add(get_task_results_shape(sub_task.task))
                    if len(sub_task_shapes) > 1:
                        msg = 'The outputs of the sub-tasks have different shapes.'
                        task_warnings[task]['other'].append([msg])

        # Ranges of repeated tasks
        # - Functional ranges
        #   - Reference other ranges
        #   - Parameters of functional ranges
        #     - Have ids
        #   - Variables of functional ranges
        #     - Have ids
        #     - Reference models
        #     - Have symbols or targets
        #   - Have math
        #   - Graph is acyclic
        for task in doc.tasks:
            if isinstance(task, RepeatedTask):
                if not isinstance(task.range, Range):
                    task_errors[task]['ranges'].append(['Repeated task must have a main range.'])

                for i_range, range in enumerate(task.ranges):
                    range_errors = []
                    range_warnings = []

                    if not range or not range.id:
                        range_errors.append(['Range must have an id.'])

                    if isinstance(range, FunctionalRange):
                        if not range.range:
                            range_errors.append(['Functional range must reference another range.'])

                        for i_param, param in enumerate(range.parameters):
                            if not param.id:
                                range_errors.append(['Parameter {} must have an id.'.format(i_param + 1)])

                        for i_variable, variable in enumerate(range.variables):
                            variable_errors = []
                            variable_warnings = []

                            if not variable.id:
                                variable_errors.append(['Variable must have an id.'])

                            if not variable.model:
                                variable_errors.append(['Variable must reference a model.'])
                            if variable.task:
                                variable_errors.append(['Variable should not reference a task.'])

                            if not variable.symbol and not variable.target:
                                variable_errors.append(['Variable should define a symbol or target.'])
                            if variable.symbol and variable.target:
                                variable_errors.append(['Variable should define a symbol or target, not both.'])

                            if variable.target and variable.model and variable.model.language:
                                temp_errors, temp_warnings = validate_target(
                                    variable.target, variable.target_namespaces, Calculation,
                                    variable.model.language, variable.model.id,
                                    check_in_model_source=(
                                        not variable.model.changes
                                        and not get_model_changes_for_task(task)
                                    )
                                )
                                variable_errors.extend(temp_errors)
                                variable_warnings.extend(temp_warnings)

                            if variable_errors:
                                variable_id = '`' + variable.id + '`' if variable and variable.id else str(i_variable + 1)
                                range_errors.append(['Variable {} is invalid.'.format(variable_id), variable_errors])

                            if variable_warnings:
                                variable_id = '`' + variable.id + '`' if variable and variable.id else str(i_variable + 1)
                                range_warnings.append(['Variable {} may be invalid.'.format(variable_id), variable_warnings])

                        temp_errors, temp_warnings = validate_calculation(range)
                        range_errors.extend(temp_errors)
                        range_warnings.extend(temp_warnings)

                    elif isinstance(range, UniformRange):
                        range_errors.extend(validate_uniform_range(range))

                    elif isinstance(range, VectorRange):
                        pass

                    else:
                        range_errors.append(['Range must be a functional, uniform, or vector range, not an instance of `{}`.'.format(
                            range.__class__.__name__)])

                    if range_errors:
                        range_id = '`' + range.id + '`' if range and range.id else str(i_range + 1)
                        task_errors[task]['ranges'].append(['{} {} is invalid.'.format(range.__class__.__name__, range_id), range_errors])

                    if range_warnings:
                        range_id = '`' + range.id + '`' if range and range.id else str(i_range + 1)
                        task_warnings[task]['ranges'].append(
                            ['{} {} may be invalid.'.format(range.__class__.__name__, range_id), range_warnings])

        # ranges of repeated tasks have unique ids
        ranges_have_ids = True
        range_ids = []
        for task in doc.tasks:
            if isinstance(task, RepeatedTask):
                for i_range, range in enumerate(task.ranges):
                    if isinstance(range, Range) and range.id:
                        range_ids.append(range.id)
                    else:
                        ranges_have_ids = False

        duplicate_range_ids = [item for item, count in collections.Counter(range_ids).items() if count > 1]
        if duplicate_range_ids:
            msg = ('Ranges must have unique ids. The following range ids are repeated:\n  - {}').format(
                '\n  - '.join(sorted(duplicate_range_ids)))
            task_errors[task]['ranges'].append([msg])

        # graph of ranges is acyclic
        check_range_cycles = ranges_have_ids and not duplicate_range_ids
        functional_range_graph = networkx.DiGraph()
        for task in doc.tasks:
            if isinstance(task, RepeatedTask):
                for range in task.ranges:
                    if isinstance(range, FunctionalRange):
                        if range.id and range.range and range.range.id:
                            if range.id:
                                functional_range_graph.add_node(range.id)
                                if range.range.id:
                                    functional_range_graph.add_edge(range.id, range.range.id)
                        else:
                            check_range_cycles = False

        check_range_lens = False
        if check_range_cycles:
            try:
                networkx.algorithms.cycles.find_cycle(functional_range_graph)
                subtasks_cyclic = True
                errors.append(['The functional ranges are defined cyclically. The graph of functional ranges must be acyclic.'])
            except networkx.NetworkXNoCycle:
                check_range_lens = True

        # ranges of repeated tasks
        # - are of a supported type (error)
        # - are at least as long as the main range (error)
        # - the same lengths (warnings)
        if check_range_lens:
            for task in doc.tasks:
                if isinstance(task, RepeatedTask):
                    if isinstance(task.range, Range):
                        main_range_len = get_range_len(task.range)
                    else:
                        main_range_len = 0

                    for range in task.ranges:
                        range_len = get_range_len(range)
                        if range_len < main_range_len:
                            msg = (
                                'The child ranges of repeated tasks must be at least as long as '
                                'the main range of their parent repeated tasks. '
                                'Range `{}` is shorter than its main range, {} < {}.'
                            ).format(range.id, range_len, main_range_len)
                            task_errors[task]['ranges'].append([msg])
                        elif range_len > main_range_len:
                            msg = ('Child range `{}` is longer than its main range ({} > {}). '
                                   'The tail elements of the range will be ignored.').format(range.id, range_len, main_range_len)
                            task_warnings[task]['ranges'].append([msg])

                    for i_change, change in enumerate(task.changes):
                        if change.range:
                            range_len = get_range_len(change.range)
                            if range_len < main_range_len:
                                msg = (
                                    'The child ranges of repeated tasks must be '
                                    'at least as long as the main range of their parent repeated tasks. '
                                    'Range `{}` is shorter than its main range, {} < {}.'
                                ).format(change.range.id, range_len, main_range_len)
                                task_errors[task]['ranges'].append([msg])
                            elif range_len > main_range_len:
                                msg = (
                                    'Child range `{}` is longer than its main range ({} > {}). '
                                    'The tail elements of the range will be ignored.'
                                ).format(change.range.id, range_len, main_range_len)
                                task_warnings[task]['ranges'].append([msg])

        # Set value changes of repeated tasks
        # - Changes reference models
        # - Changes reference a symbol or target
        # - Parameters
        #   - Have ids
        # - Variables
        #   - Have ids
        #   - Reference a model, not a task
        #   - Define a target, not a symbol
        # - Changes have math
        for task in doc.tasks:
            if isinstance(task, RepeatedTask):
                all_change_errors = []
                all_change_warnings = []
                for i_change, change in enumerate(task.changes):
                    change_errors = []
                    change_warnings = []

                    if not change.model:
                        msg = ('Set value change must reference a model. '
                               'Change does not reference a model.')
                        change_errors.append([msg])

                    if change.target:
                        if change.model and change.model.language:
                            temp_errors, temp_warnings = validate_target(
                                change.target, change.target_namespaces,
                                ModelChange, change.model.language, change.model.id,
                                check_in_model_source=False)
                            change_errors.extend(temp_errors)
                            change_warnings.extend(temp_warnings)

                    else:
                        msg = ('Set value changes must define a target. '
                               'Change does not define a target.')
                        change_errors.append([msg])

                    if change.symbol:
                        msg = ('Set value changes should not define a symbol. '
                               'Change defines a symbol.')
                        change_errors.append([msg])

                    for i_parameter, parameter in enumerate(change.parameters):
                        if not parameter.id:
                            change_errors.append(['Parameter {} must have an id.'.format(i_parameter + 1)])

                    for i_variable, variable in enumerate(change.variables):
                        variable_errors = []
                        variable_warnings = []

                        if not variable.id:
                            variable_errors.append(['Variable must have an id.'])

                        if not variable.model:
                            variable_errors.append(['Set value variable must reference a model.'])
                        if variable.task:
                            variable_errors.append(['Set value variable should not reference a task.'])

                        if not variable.target and not variable.symbol:
                            variable_errors.append(['Set value variable must define a target or a symbol.'])
                        if variable.target and variable.symbol:
                            variable_errors.append(['Set value variable must define a target or a symbol, not both.'])

                        if variable.target and variable.model and variable.model.language:
                            temp_errors, temp_warnings = validate_target(
                                variable.target, variable.target_namespaces,
                                Calculation, variable.model.language, variable.model.id,
                                check_in_model_source=False)
                            variable_errors.extend(temp_errors)
                            variable_warnings.extend(temp_warnings)

                        if variable_errors:
                            variable_id = '`' + variable.id + '`' if variable and variable.id else str(i_variable + 1)
                            change_errors.append(['Variable {} is invalid.'.format(variable_id), variable_errors])

                        if variable_warnings:
                            variable_id = '`' + variable.id + '`' if variable and variable.id else str(i_variable + 1)
                            change_warnings.append(['Variable {} may be invalid.'.format(variable_id), variable_warnings])

                    temp_errors, temp_warnings = validate_calculation(change)
                    change_errors.extend(temp_errors)
                    change_warnings.extend(temp_warnings)

                    if change_errors:
                        change_id = '`' + change.id + '`' if change and change.id else str(i_change + 1)
                        all_change_errors.append(['Change {} is invalid.'.format(change_id), change_errors])

                    if change_warnings:
                        change_id = '`' + change.id + '`' if change and change.id else str(i_change + 1)
                        all_change_warnings.append(['Change {} may be invalid.'.format(change_id), change_warnings])

                if all_change_errors:
                    task_errors[task]['other'].append(['Changes are invalid.', all_change_errors])

                if all_change_warnings:
                    task_warnings[task]['other'].append(['Changes may be invalid.', all_change_warnings])

        # repeated tasks involve 1 model
        if not subtasks_cyclic:
            for task in doc.tasks:
                if isinstance(task, RepeatedTask):
                    temp_errors, temp_warnings = validate_repeated_task_has_one_model(task)
                    task_errors[task]['other'].extend(temp_errors)
                    task_warnings[task]['other'].extend(temp_warnings)

        for i_task, task in enumerate(doc.tasks):
            task_id = '`' + task.id + '`' if task and task.id else str(i_task + 1)

            task_errors[task]['other'] += task_errors[task]['ranges']
            task_warnings[task]['other'] += task_warnings[task]['ranges']

            if task_errors[task]['other']:
                errors.append(['Task {} is invalid.'.format(task_id), task_errors[task]['other']])

            if task_warnings[task]['other']:
                warnings.append(['Task {} may be invalid.'.format(task_id), task_warnings[task]['other']])

        # validate data generators
        if validate_targets_with_model_sources:
            model_etrees = {}
            for model in doc.models:
                if (
                    model.language
                    and is_model_language_encoded_in_xml(model.language)
                    and model.source
                    and not model.source.startswith('#')
                    and not model.source.startswith('urn:')
                    and not model.source.startswith('http://')
                    and not model.source.startswith('https://')
                ):
                    if os.path.isabs(model.source):
                        model_source = model.source
                    else:
                        model_source = os.path.join(working_dir, model.source)

                    if os.path.isfile(model_source):
                        try:
                            model_etrees[model] = lxml.etree.parse(model_source)
                        except Exception:
                            pass
        else:
            model_etrees = None

        for i_data_gen, data_gen in enumerate(doc.data_generators):
            data_gen_errors, data_gen_warnings = validate_data_generator(
                data_gen, model_etrees=model_etrees, validate_targets_with_model_sources=validate_targets_with_model_sources)

            data_gen_id = '`' + data_gen.id + '`' if data_gen and data_gen.id else str(i_data_gen + 1)

            if data_gen_errors:
                errors.append(['Data generator {} is invalid.'.format(data_gen_id), data_gen_errors])

            if data_gen_warnings:
                warnings.append(['Data generator {} may be invalid.'.format(data_gen_id), data_gen_warnings])

        # validate outputs
        for i_output, output in enumerate(doc.outputs):
            output_errors, output_warnings = validate_output(output)

            output_id = '`' + output.id + '`' if output and output.id else str(i_output + 1)

            if output_errors:
                errors.append(['Output {} is invalid.'.format(output_id), output_errors])

            if output_warnings:
                warnings.append(['Output {} may be invalid.'.format(output_id), output_warnings])

        # tasks, data generators that don't contribute to outputs
        used_data_generators = set()
        used_tasks = set()
        for output in doc.outputs:
            if isinstance(output, Report):
                for data_set in output.data_sets:
                    if data_set.data_generator:
                        used_data_generators.add(data_set.data_generator)
                        for variable in data_set.data_generator.variables:
                            if variable.task:
                                used_tasks.add(variable.task)

            elif isinstance(output, Plot2D):
                for curve in output.curves:
                    if curve.x_data_generator:
                        used_data_generators.add(curve.x_data_generator)
                        for variable in curve.x_data_generator.variables:
                            if variable.task:
                                used_tasks.add(variable.task)
                    if curve.y_data_generator:
                        used_data_generators.add(curve.y_data_generator)
                        for variable in curve.y_data_generator.variables:
                            if variable.task:
                                used_tasks.add(variable.task)

            elif isinstance(output, Plot3D):
                for surface in output.surfaces:
                    if surface.x_data_generator:
                        used_data_generators.add(surface.x_data_generator)
                        for variable in surface.x_data_generator.variables:
                            if variable.task:
                                used_tasks.add(variable.task)
                    if surface.y_data_generator:
                        used_data_generators.add(surface.y_data_generator)
                        for variable in surface.y_data_generator.variables:
                            if variable.task:
                                used_tasks.add(variable.task)
                    if surface.z_data_generator:
                        used_data_generators.add(surface.z_data_generator)
                        for variable in surface.z_data_generator.variables:
                            if variable.task:
                                used_tasks.add(variable.task)

        tasks_to_check = list(used_tasks)
        used_tasks = set()
        while tasks_to_check:
            task = tasks_to_check.pop()
            used_tasks.add(task)
            if isinstance(task, RepeatedTask):
                for sub_task in task.sub_tasks:
                    tasks_to_check.append(sub_task.task)

        unused_tasks = []
        for i_task, task in enumerate(doc.tasks):
            if task not in used_tasks:
                task_id = '`' + task.id + '`' if task.id else str(i_task + 1)
                unused_tasks.append([task_id])
        if unused_tasks:
            warnings.append(['The following tasks do not contribute to outputs:', sorted(unused_tasks)])

        unused_data_generators = []
        for i_data_generator, data_generator in enumerate(doc.data_generators):
            if data_generator not in used_data_generators:
                data_generator_id = '`' + data_generator.id + '`' if data_generator.id else str(i_data_generator + 1)
                unused_data_generators.append([data_generator_id])
        if unused_data_generators:
            warnings.append(['The following data generators do not contribute to outputs:', sorted(unused_data_generators)])

    return (errors, warnings)


def validate_reference(obj, obj_label, attr_name, attr_label):
    errors = []

    if not getattr(obj, attr_name):
        errors.append(['{} `{}` must have a {}.'.format(obj_label, obj.id, attr_label)])

    return errors


def validate_unique_ids(doc):
    """ Validate that the element of a SED document have unique ids

    Args:
        doc (:obj:`SedDocument`): SED document

    Returns:
        nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []
    objs = get_all_sed_objects(doc, SedIdGroupMixin)

    ids = set()
    duplicate_ids = set()
    for obj in objs:
        id = obj.id
        if id:
            if id in ids:
                duplicate_ids.add(id)
            ids.add(id)

    if duplicate_ids:
        errors.append(['Each identified SED object must have a unique id. Multiple objects have the following ids:',
                       [[id] for id in sorted(duplicate_ids)]])

    return errors


def validate_task(task):
    """ Validate a simulation task

    Args:
        task (:obj:`Task`): task

    Returns:
        nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []

    if isinstance(task, Task):

        # check that task has model
        if not task.model:
            errors.append(['Task must have a model.'])

        # check that task has simulation
        if not task.simulation:
            errors.append(['Task must have a simulation.'])

    else:
        errors.append(['Task must be an instance of `Task` not `{}`.'.format(task.__class__.__name__)])

    # return errors
    return errors


def validate_repeated_task_has_one_model(task):
    """ Validate a that a repeated task involves a single model

    Args:
        task (:obj:`RepeatedTask`): repeated task

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of warnings (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []
    models = get_models_referenced_by_task(task)
    if len(models) > 1:
        msg = (
            'The use of multiple models with a repeated task is a developmental feature of SED-ML. '
            'This task may not be executed consistently across simulation tools.'
        )
        warnings.append([msg])
    return (errors, warnings)


def validate_model(model, model_ids, working_dir, validate_models_with_languages=True):
    """ Check a model

    Args:
        model (:obj:`Model`): model
        model_ids (:obj:`list` of :obj:`str`): ids of models
        working_dir (:obj:`str`): working directory (e.g., for referencing model files)
        validate_models_with_languages (:obj:`bool`, optional): if :obj:`True`, validate models

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of warnings (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    # check source
    if not model.language:
        errors.append(['Model must have a language.'])

    # check source
    tmp_errors, tmp_warnings = validate_model_source(
        model, model_ids, working_dir, validate_models_with_languages=validate_models_with_languages)
    errors.extend(tmp_errors)
    warnings.extend(tmp_warnings)

    # validate that model changes have targets
    model_change_errors, model_change_warnings = validate_model_changes(model)
    if model_change_errors:
        errors.append(['The changes of the model are invalid.', model_change_errors])
    if model_change_warnings:
        warnings.append(['The changes of the model may be invalid.', model_change_warnings])

    return (errors, warnings)


def validate_model_language(language, valid_languages):
    """ Check that model is encoded in a specific language

    Args:
        language (:obj:`str`): model language
        valid_language (:obj:`ModelLanguage` or :obj:`list` of :obj:`ModelLanguage`): valid language(s)

    Returns:
        nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []

    if not isinstance(valid_languages, (list, tuple)):
        valid_languages = [valid_languages]

    is_valid = False
    for valid_language in valid_languages:
        valid_language_pattern = ModelLanguagePattern[valid_language.name]

        if language and re.match(valid_language_pattern.value, language):
            is_valid = True
            break

    if not is_valid:
        msg = (
            "Model language `{}` is not supported. "
            "Models must be in {} format (e.g., `sed:model/@language` must match `{}` such as `{}`)."
        ).format(language or '',
                 ', '.join(valid_language.name for valid_language in valid_languages),
                 valid_language_pattern.value,
                 valid_language.value)
        errors.append([msg])

    return errors


def validate_model_source(model, model_ids, working_dir, validate_models_with_languages=True):
    """ Check the source of a model

    Args:
        model (:obj:`Model`): model
        model_ids (:obj:`list` of :obj:`str`): ids of models
        working_dir (:obj:`str`): working directory (e.g., for referencing model files)
        validate_models_with_languages (:obj:`bool`, optional): if :obj:`True`, validate models

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of warnings (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    if not model.source:
        errors.append(['Model must have a source.'])

    elif model.source.startswith('#'):
        if model.source[1:] not in model_ids:
            errors.append(['The referenced source `{}` is not defined.'.format(model.source)])

    elif model.source.startswith('urn:'):
        warnings.append([(
            'Model resolution via URNs will likely be deprecated in a future version of SED-ML. '
            'Instead models can be resolved by URLs.'
        )])
        warnings.append(['URN model source `{}` was not validated.'.format(model.source)])

    elif model.source.startswith('http://') or model.source.startswith('https://'):
        warnings.append(['URL model source `{}` was not validated.'.format(model.source)])

    else:
        if os.path.isabs(model.source):
            model_source = model.source
        else:
            model_source = os.path.join(working_dir, model.source)

        if validate_models_with_languages:
            model_source_errors, model_source_warnings, _ = validate_model_with_language(model_source, model.language, name=model.id)
            if model_source_errors:
                errors.append(['The model file `{}` is invalid.'.format(model.source), model_source_errors])
            if model_source_warnings:
                warnings.append(['The model file `{}` may be invalid.'.format(model.source), model_source_warnings])

    return (errors, warnings)


def validate_model_with_language(source, language, name=None):
    """ Check that a model is valid

    Args:
        source (:obj:`str`): path to model
        language (:obj:`ModelLanguage`): language
        name (:obj:`str`, optional): name of model for use in error messages

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`object`: model object
    """
    errors = []
    warnings = []

    if language and re.match(ModelLanguagePattern.BNGL, language):
        from ..model_lang.bngl.validation import validate_model

    elif language and re.match(ModelLanguagePattern.CellML, language):
        from ..model_lang.cellml.validation import validate_model

    elif language and re.match(ModelLanguagePattern.LEMS, language):
        from ..model_lang.lems.validation import validate_model

    elif language and re.match(ModelLanguagePattern.NeuroML, language):
        from ..model_lang.neuroml.validation import validate_model

    elif language and re.match(ModelLanguagePattern.RBA, language):
        from ..model_lang.rba.validation import validate_model

    elif language and re.match(ModelLanguagePattern.SBML, language):
        from ..model_lang.sbml.validation import validate_model

    elif language and re.match(ModelLanguagePattern.Smoldyn, language):
        from ..model_lang.smoldyn.validation import validate_model

    elif language and re.match(ModelLanguagePattern.XPP, language):
        from ..model_lang.xpp.validation import validate_model

    else:
        warnings.append(['No validation is available for models encoded in `{}`'.format(getattr(language, 'name', language) or '')])
        return (errors, warnings, None)

    return validate_model(source, name=name)


def validate_model_change_types(changes, types=(ModelChange, )):
    """ Check that model changes are valid

    Args:
        changes (:obj:`list` of :obj:`ModelChange`): model changes
        types (:obj:`type` or :obj:`tuple` of :obj:`type`, optional): valid type(s) of model changes

    Returns:
        nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []

    for i_change, change in enumerate(changes):
        if not isinstance(change, types):
            msg = "".join([
                'Model change {} of type `{}` is not supported. '.format(i_change + 1, change.__class__.__name__),
                'Model changes must be instances of one of the following types:\n  - {}'.format(
                    '\n  - '.join(type.__name__ for type in types)),
            ])
            errors.append([msg])

    return errors


def validate_model_changes(model):
    """ Check that model changes are semantically valid

    * Check that the variables of compute model changes are valid

    Args:
        model (:obj:`Model`): model

    Returns:
        nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    for i_change, change in enumerate(model.changes):
        change_errors = []
        change_warnings = []

        if change.target:
            if model.language:
                temp_errors, temp_warnings = validate_target(change.target, change.target_namespaces,
                                                             ModelChange, model.language, model.id,
                                                             check_in_model_source=False)
                change_errors.extend(temp_errors)
                change_warnings.extend(temp_warnings)

        else:
            change_errors.append(['Model attribute change must define a target.'])

        if isinstance(change, ComputeModelChange):
            for i_parameter, parameter in enumerate(change.parameters):
                if not parameter.id:
                    change_errors.append(['Parameter {} must have an id.'.format(i_parameter + 1)])

            for i_variable, variable in enumerate(change.variables):
                variable_errors = []
                variable_warnings = []

                if not variable.id:
                    variable_errors.append(['Variable must have an id.'])

                if not variable.model:
                    variable_errors.append(['Variable must reference a model.'])
                if variable.task:
                    variable_errors.append(['Variable should not reference a task.'])

                if not variable.target:
                    variable_errors.append(['Variable must define a target.'])
                if variable.symbol:
                    variable_errors.append(['Variable must define a target, not a symbol.'])

                if variable.target and variable.model and variable.model.language:
                    temp_errors, temp_warnings = validate_target(
                        variable.target, variable.target_namespaces,
                        Calculation, variable.model.language, variable.model.id,
                        check_in_model_source=False)
                    variable_errors.extend(temp_errors)
                    variable_warnings.extend(temp_warnings)

                if variable_errors:
                    var_id = '`' + variable.id + '`' if variable and variable.id else str(i_variable + 1)
                    change_errors.append(['Variable {} is invalid.'.format(var_id), variable_errors])

                if variable_warnings:
                    var_id = '`' + variable.id + '`' if variable and variable.id else str(i_variable + 1)
                    change_warnings.append(['Variable {} may be invalid.'.format(var_id), variable_warnings])

            temp_errors, temp_warnings = validate_calculation(change)
            change_errors.extend(temp_errors)
            change_warnings.extend(temp_warnings)

        if change_errors:
            change_id = '`' + change.id + '`' if change and change.id else str(i_change + 1)
            errors.append(['Change {} is invalid.'.format(change_id), change_errors])

        if change_warnings:
            change_id = '`' + change.id + '`' if change and change.id else str(i_change + 1)
            warnings.append(['Change {} may be invalid.'.format(change_id), change_warnings])

    return errors, warnings


def validate_simulation_type(simulation, types):
    """ Check that simulation is a time course simulation

    Args:
        simulation (:obj:`Simulation`): simulation
        types (:obj:`type`): valid simulation types

    Returns:
        nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []

    if not isinstance(simulation, types):
        errors.append([
            'Simulation {} of type `{}` is not supported. Simulation must be an instance of one of the following:\n  - {}'.format(
                simulation.id, simulation.__class__.__name__, '\n  - '.join(type.__name__ for type in types))
        ])

    return errors


def validate_simulation(simulation):
    """ Check that simulation is a valid simulation

    Args:
        simulation (:obj:`Simulation`): simulation

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    if isinstance(simulation, OneStepSimulation):
        if not isinstance(simulation.step, (int, float)):
            errors.append(['Step must be a float.'])

    elif isinstance(simulation, SteadyStateSimulation):
        pass

    elif isinstance(simulation, UniformTimeCourseSimulation):
        if simulation.initial_time is None:
            errors.append(['Inital time must be set.'])
        if simulation.output_start_time is None:
            errors.append(['Output start time must be set.'])
        if simulation.output_end_time is None:
            errors.append(['Output end time must be set.'])
        if simulation.number_of_steps is None:
            errors.append(['Number of steps must be set.'])

        if (
            simulation.initial_time is not None
            and simulation.output_start_time is not None
            and simulation.output_start_time < simulation.initial_time
        ):
            errors.append(['Output start time {} must be at least the initial time {}.'.format(
                simulation.output_start_time, simulation.initial_time)])

        if (
            simulation.output_start_time is not None
            and simulation.output_end_time is not None
            and simulation.output_end_time < simulation.output_start_time
        ):
            errors.append(['Output end time {} must be at least the output start time {}.'.format(
                simulation.output_end_time, simulation.output_start_time)])

        if simulation.number_of_steps is not None:
            if simulation.number_of_steps < 1:
                errors.append(['Number of points must be at least 1.'])

            if math.floor(simulation.number_of_steps) != simulation.number_of_steps:
                errors.append(['Number of points must be an integer.'])

            elif (simulation.number_of_steps % 5) != 0:
                warnings.append([
                    'The time course simulation has an unusual number of steps. Check that the parameters of the time course are correct.',
                    [
                        ['Initial time: {}'.format(simulation.initial_time if simulation.initial_time is not None else '')],
                        ['Output start time: {}'.format(simulation.output_start_time if simulation.output_start_time is not None else '')],
                        ['Output end time: {}'.format(simulation.output_end_time if simulation.output_end_time is not None else '')],
                        ['Number of steps: {}'.format(simulation.number_of_steps if simulation.number_of_steps is not None else '')],
                    ],
                ])

    else:
        errors.append(['Simulation must be a one step, steady-state, or uniform time course, not an instance of `{}`.'.format(
            simulation.__class__.__name__)])

    if simulation.algorithm:
        temp_errors, temp_warnings = validate_algorithm(simulation.algorithm)
        errors.extend(temp_errors)
        warnings.extend(temp_warnings)

    else:
        errors.append(['Simulation must have an algorithm.'])

    return errors, warnings


def validate_algorithm(algorithm):
    """ Validate an algorithm

    Args:
        algorithm (:obj:`Algorithm`): algorithm

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    if get_kisao_term_type(get_kisao_term(algorithm.kisao_id)) != KisaoTermType.algorithm:
        errors.append(['Algorithm has an invalid KiSAO id `{}`.'.format(algorithm.kisao_id)])

    change_kisao_ids = []
    for i_change, change in enumerate(algorithm.changes):
        if get_kisao_term_type(get_kisao_term(change.kisao_id)) == KisaoTermType.algorithm_parameter:
            change_kisao_ids.append(change.kisao_id)
        else:
            errors.append(
                ['Algorithm change {} has an invalid KiSAO id `{}`.'.format(i_change + 1, change.kisao_id)])

    unique_change_kisao_ids = set(change_kisao_ids)
    if len(change_kisao_ids) > len(unique_change_kisao_ids):
        duplicate_change_kisao_ids = sorted([id] for id in unique_change_kisao_ids if change_kisao_ids.count(id) > 1)
        errors.append(['Each algorithm parameter must have a unique KiSAO id. Multiple parameters have the following KiSAO ids:',
                       duplicate_change_kisao_ids])

    return errors, warnings


def validate_uniform_range(range):
    """ Validate a uniform range

    Args:
        range (:obj:`data_model.UniformRange`): range

    Returns:
        nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []

    if range.number_of_steps < 1:
        errors.append(['Uniform range `{}` must have at least one step.'.format(
            range.id)])

    return errors


def validate_data_generator(data_generator, model_etrees=None, validate_targets_with_model_sources=True):
    """ Validate a data generator

    * Parameters have ids
    * Variables have ids
    * Variables have target OR symbol
    * Variables don't have model references
    * Have math
    * Math is valid

    Args:
        data_generator (:obj:`DataGenerator`): data generator
        model_etrees (:obj:`dict`, optional): dictionary that maps models to XML element trees of their sources
        validate_targets_with_model_sources (:obj:`bool`, optional): whether to validate the targets of the variables of the data generator

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of warnings (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    for i_parameter, param in enumerate(data_generator.parameters):
        if not param.id:
            errors.append(['Parameter {} must have an id.'.format(i_parameter + 1)])

    temp_errors, temp_warnings = validate_data_generator_variables(data_generator.variables, model_etrees=model_etrees,
                                                                   validate_targets_with_model_sources=validate_targets_with_model_sources)
    errors.extend(temp_errors)
    warnings.extend(temp_warnings)

    temp_errors, temp_warnings = validate_calculation(data_generator)
    errors.extend(temp_errors)
    warnings.extend(temp_warnings)

    return (errors, warnings)


def validate_data_generator_variables(variables, model_etrees=None, validate_targets_with_model_sources=True):
    """ Check variables have a symbol or target

    Args:
        variables (:obj:`list` of :obj:`Variable`): variables
        model_etrees (:obj:`dict`, optional): dictionary that maps models to XML element trees of their sources
        validate_targets_with_model_sources (:obj:`bool`, optional): whether to validate the targets of the variables of the data generator

    Returns:
        nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    model_etrees = model_etrees or {}

    errors = []
    warnings = []

    task_types = set()

    for i_variable, variable in enumerate(variables):
        variable_errors = []
        variable_warnings = []

        if not variable.id:
            variable_errors.append(['Variable must have an id.'])

        if variable.model:
            variable_errors.append(['Variable should not reference a model.'])

        if variable.task:
            task_types.add(get_task_results_shape(variable.task))
        elif not (variable.target and variable.target.startswith('#')):
            variable_errors.append(['Variable must reference a task.'])

        if (variable.symbol and variable.target) or (not variable.symbol and not variable.target):
            variable_errors.append(['Variable must define a symbol or target.'])

        if variable.target and variable.task:
            models = get_models_referenced_by_task(variable.task)
            for model in models:
                if model and model.language:
                    model_changes = model.changes or list(filter(lambda change: change.model == model,
                                                                 get_model_changes_for_task(variable.task)))

                    temp_errors, temp_warnings = validate_target(
                        variable.target, variable.target_namespaces,
                        DataGenerator, model.language,
                        model_id=model.id,
                        model_etree=model_etrees.get(model, None),
                        check_in_model_source=not model_changes and validate_targets_with_model_sources)
                    variable_errors.extend(temp_errors)
                    variable_warnings.extend(temp_warnings)

        if variable_errors:
            variable_id = '`' + variable.id + '`' if variable and variable.id else str(i_variable + 1)
            errors.append(['Variable {} is invalid.'.format(variable_id), variable_errors])

        if variable_warnings:
            variable_id = '`' + variable.id + '`' if variable and variable.id else str(i_variable + 1)
            warnings.append(['Variable {} may be invalid.'.format(variable_id), variable_warnings])

    if len(task_types) > 1:
        warnings.append(['The variables do not have consistent shapes.'])

    return errors, warnings


def validate_output(output):
    """ Validate an output

    * Reports
      * Data sets have ids and labels
      * Data sets reference data generators

    * Plots

       * Curves and surfaces have ids
       * x, y, z attributes reference data generators

    Args:
        output (:obj:`Output`): output

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of warnings (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    if isinstance(output, Report):
        if not output.data_sets:
            errors.append(['Report must have at least one data set.'])

        labels = set()
        duplicate_labels = set()
        task_types = set()

        for i_data_set, data_set in enumerate(output.data_sets):
            data_set_errors = []

            if not data_set.id:
                data_set_errors.append(['Data set must have an id.'])

            if data_set.label:
                if data_set.label in labels:
                    duplicate_labels.add(data_set.label)
                labels.add(data_set.label)
            else:
                data_set_errors.append(['Data set must have a label.'])

            if data_set.data_generator:
                for variable in data_set.data_generator.variables:
                    if variable.task:
                        task_types.add(get_task_results_shape(variable.task))
            data_set_errors.extend(validate_reference(data_set, 'Data set', 'data_generator', 'data data generator'))

            if data_set_errors:
                data_set_id = '`' + data_set.id + '`' if data_set and data_set.id else str(i_data_set + 1)
                errors.append(['Data set {} is invalid.'.format(data_set_id), data_set_errors])

        if len(task_types) > 1:
            warnings.append(['The data sets do not have consistent shapes.'])

        if duplicate_labels:
            warnings.append([(
                'Data sets do not have unique labels. '
                'Unique labels are helpful for interpreting reports. '
                'The following labels are repeated:'),
                [[label] for label in sorted(duplicate_labels)]])

    elif isinstance(output, Plot2D):
        if not output.curves:
            errors.append(['Plot must have at least one curve.'])

        x_scales = set()
        y_scales = set()

        for i_curve, curve in enumerate(output.curves):
            curve_errors = []
            curve_warnings = []

            if not curve.id:
                curve_errors.append(['Curve must have an id.'])

            task_types = set()
            if curve.x_data_generator:
                for variable in curve.x_data_generator.variables:
                    if variable.task:
                        task_types.add(get_task_results_shape(variable.task))
            if curve.y_data_generator:
                for variable in curve.y_data_generator.variables:
                    if variable.task:
                        task_types.add(get_task_results_shape(variable.task))
            curve_errors.extend(validate_reference(curve, 'Curve', 'x_data_generator', 'x data data generator'))
            curve_errors.extend(validate_reference(curve, 'Curve', 'y_data_generator', 'y data data generator'))
            if len(task_types) > 1:
                curve_warnings.append(['The curves do not have consistent shapes.'])

            if curve.x_scale:
                x_scales.add(curve.x_scale)
            else:
                curve_errors.append(['Curve must have an x-scale.'])

            if curve.y_scale:
                y_scales.add(curve.y_scale)
            else:
                curve_errors.append(['Curve must have a y-scale.'])

            if curve_errors:
                curve_id = '`' + curve.id + '`' if curve and curve.id else str(i_curve + 1)
                errors.append(['Curve {} is invalid.'.format(curve_id), curve_errors])
            if curve_warnings:
                curve_id = '`' + curve.id + '`' if curve and curve.id else str(i_curve + 1)
                warnings.append(['Curve {} may be invalid.'.format(curve_id), curve_warnings])

        if len(x_scales) > 1:
            warnings.append(['Curves do not have consistent x-scales.'])
        if len(y_scales) > 1:
            warnings.append(['Curves do not have consistent y-scales.'])

    elif isinstance(output, Plot3D):
        if not output.surfaces:
            errors.append(['Plot must have at least one surface.'])

        x_scales = set()
        y_scales = set()
        z_scales = set()

        for i_surface, surface in enumerate(output.surfaces):
            surface_errors = []
            surface_warnings = []

            if not surface.id:
                surface_errors.append(['Surface must have an id.'])

            task_types = set()
            if surface.x_data_generator:
                for variable in surface.x_data_generator.variables:
                    if variable.task:
                        task_types.add(get_task_results_shape(variable.task))
            if surface.y_data_generator:
                for variable in surface.y_data_generator.variables:
                    if variable.task:
                        task_types.add(get_task_results_shape(variable.task))
            if surface.z_data_generator:
                for variable in surface.z_data_generator.variables:
                    if variable.task:
                        task_types.add(get_task_results_shape(variable.task))
            surface_errors.extend(validate_reference(surface, 'Surface', 'x_data_generator', 'x data data generator'))
            surface_errors.extend(validate_reference(surface, 'Surface', 'y_data_generator', 'y data data generator'))
            surface_errors.extend(validate_reference(surface, 'Surface', 'z_data_generator', 'z data data generator'))
            if len(task_types) > 1:
                surface_warnings.append(['The surfaces do not have consistent shapes.'])

            if surface.x_scale:
                x_scales.add(surface.x_scale)
            else:
                surface_errors.append(['Surface must have an x-scale.'])

            if surface.y_scale:
                y_scales.add(surface.y_scale)
            else:
                surface_errors.append(['Surface must have a y-scale.'])

            if surface.z_scale:
                z_scales.add(surface.z_scale)
            else:
                surface_errors.append(['Surface must have a z-scale.'])

            if surface_errors:
                surface_id = '`' + surface.id + '`' if surface and surface.id else str(i_surface + 1)
                errors.append(['Surface {} is invalid.'.format(surface_id), surface_errors])
            if surface_warnings:
                surface_id = '`' + surface.id + '`' if surface and surface.id else str(i_surface + 1)
                warnings.append(['Surface {} may be invalid.'.format(surface_id), surface_warnings])

        if len(x_scales) > 1:
            warnings.append(['Surfaces do not have consistent x-scales.'])
        if len(y_scales) > 1:
            warnings.append(['Surfaces do not have consistent y-scales.'])
        if len(z_scales) > 1:
            warnings.append(['Surfaces do not have consistent z-scales.'])

    involves_repeated_task = False
    for variable in get_variables_for_data_generators(get_data_generators_for_output(output)):
        if variable.task and not isinstance(variable.task, Task):
            involves_repeated_task = True

    if involves_repeated_task:
        msg = (
            'Some simulation tools will not be able to generate this output because it uses data from repeated tasks. '
            'Output for repeated tasks is an experimental feature of SED-ML, which is not officially supported.'
        )
        warnings.append([msg])

    return (errors, warnings)


def validate_target(target, namespaces, context, language, model_id, model_etree=None, doc=None, check_in_model_source=False):
    """ Validate that a target is a valid XPath and that the namespaces needed to resolve a target are defined

    Args:
        target (:obj:`str`): XPath to a model element or attribute
        namespaces (:obj:`dict`): dictionary that maps prefixes of namespaces to their URIs
        context (:obj:`type`)
        language (:obj:`str`): model language
        model_id (:obj:`str`): model id
        model_etree (:obj:`etree.Element`, optional): XML element tree for model source
        doc (:obj:`SedDocument`, optional): SED document
        check_in_model_source (:obj:`bool`, optional): whether to check that the target exists in the source

    Returns:
        nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    if target.startswith('#'):
        if context == DataGenerator:
            warnings.append(['Reference `{}` to a data descriptor could not be validated.'.format(target)])

            # TODO
            # valid_reference = False
            # for data_description in doc.data_descriptions:
            #     if data_description.id == target[1:]:
            #         valid_reference = True
            #         break
            # if not valid_reference:
            #     errors.append(['No data descriptor has an id which matches the URI fragment `{}`.'.format(target)])

        else:
            errors.append(['URI fragment targets are not supported in the context of {}.'.format(context.__name__)])

    elif does_model_language_use_xpath_variable_targets(language):

        if None in namespaces:
            namespaces = dict(namespaces)
            namespaces.pop(None, None)

        try:
            xpath = lxml.etree.XPath(target, namespaces=namespaces)
            root = lxml.etree.Element("root")
            try:
                xpath.evaluate(root)

                if model_etree and check_in_model_source and context in [DataGenerator, Calculation]:
                    if context == DataGenerator and '/@' in target:
                        xpath = target.rpartition('/@')[0]
                    else:
                        xpath = target

                    objs = eval_xpath(model_etree, xpath, namespaces)  # TODO

                    if not objs:
                        errors.append(['XPath `{}` does not match any elements of model `{}`.'.format(xpath, model_id or '')])
                    elif len(objs) > 1:
                        errors.append(['XPath `{}` matches multiple elements of model `{}`.'.format(xpath, model_id or '')])

                else:
                    warnings.append(['XPath could not be validated.'])

            except lxml.etree.XPathEvalError as exception:
                if 'Undefined namespace prefix' in str(exception):
                    if namespaces:
                        ns_message = 'Only the following namespaces are defined for the target: {}.'.format(
                            ', '.join('`' + prefix + '`' for prefix in sorted(namespaces.keys())))
                    else:
                        ns_message = 'No namespaces are defined for the target.'

                    errors.append(['One or more namespaces required for target `{}` are not defined. {}'.format(target, ns_message)])

        except lxml.etree.XPathSyntaxError:
            errors.append(['Target `{}` is not a valid XML XPath.'.format(target)])

    return errors, warnings


def validate_target_xpaths(targets, model_etree, attr='id'):
    """ Validate that the target of each model change or variable matches one object in
    an XML-encoded model and, optionally, return the value of one of its attributes

    Args:
        targets (:obj:`list` of :obj:`TargetGroupMixin`): model changes or variables
        model_source (:obj:`lxml.etree._ElementTree`): element tree for the XML model document
        attr (:obj:`str`, optional): attribute to get values of

    Returns:
        :obj:`dict` of :obj:`str` to :obj:`str`: dictionary that maps each XPath to the
            value of the attribute of the object in the XML file that matches the XPath
    """
    x_path_attrs = {}
    for target in targets:
        if target.target:
            x_path = target.target
            if '/@' in x_path:
                x_path, _, _ = x_path.rpartition('/@')
            x_path_attrs[target.target] = validate_xpaths_ref_to_unique_objects(
                model_etree, [x_path], target.target_namespaces, attr=attr)[x_path]
    return x_path_attrs


def validate_calculation(calculation):
    """ Validate that all of the symbols needed for a calculation are defined

    Args:
        expression (:obj:`Calculation`)

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of warmings (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    if calculation.math:
        workspace = {}

        for parameter in calculation.parameters:
            if parameter.id:
                workspace[parameter.id] = parameter.value

        for variable in calculation.variables:
            if variable.id:
                workspace[variable.id] = 1

        if isinstance(calculation, (FunctionalRange, SetValueComputeModelChange)):
            if calculation.range and calculation.range.id:
                workspace[calculation.range.id] = 1

        try:
            compiled_math = compile_math(calculation.math)
        except TypeError:
            errors.append(['The mathematical expression must be a `string`, not a `{}`.'.format(calculation.math.__class__)])
            return (errors, warnings)
        except SyntaxError:
            errors.append(['The syntax of the mathematical expression `{}` is invalid.'.format(calculation.math)])
            return (errors, warnings)
        except Exception as exception:
            errors.append(['The mathematical expression `{}` is invalid.'.format(calculation.math), [[str(exception)]]])
            return (errors, warnings)

        try:
            eval_math(calculation.math, compiled_math, workspace)
        except Exception as exception:
            errors.append(['The mathematical expression `{}` cannot be evaluated.'.format(calculation.math), [[str(exception)]]])
            return (errors, warnings)

    else:
        msg = 'Calculation must have math.'
        errors.append([msg])

    return (errors, warnings)



""" Standard warnings issued by biosimulation tools

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-29
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .config import Colors
import termcolor
import warnings

__all__ = [
    'BioSimulatorsWarning',
    'warn',
]


class BioSimulatorsWarning(UserWarning):
    """ Base class for simulator warnings """
    pass  # pragma: no cover


def warn(message, category):
    """ Issue a warning in a color

    Args:
        message (:obj:`str`): message
        category (:obj:`type`): category
    """
    warnings.warn(termcolor.colored(message, Colors.warning.value), category)



""" Utilities for visualizations

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-07-22
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import math

__all__ = ['cubic_bezier_point', 'cubic_bezier_angle']


def cubic_bezier_point(points, t):
    """ Get the coordinate of a point along a cubic bezier curve

    Args:
        points (:obj:`list` of :obj:`list` of :obj:`float`): control points
        t (:obj:`float`): position along the curve [0, 1]

    Returns:
        :obj:`tuple` of :obj:`float`: position of the curve at point :obj:`t`
    """
    return (
        (
            ((1. - t) ** 3.) * points[0][0]
            + 3. * t * ((1. - t) ** 2.) * points[1][0]
            + (3. * t ** 2.) * (1. - t) * points[2][0]
            + (t ** 3.) * points[3][0]
        ),
        (
            ((1. - t) ** 3.) * points[0][1]
            + 3. * t * ((1. - t) ** 2.) * points[1][1]
            + 3. * (t ** 2.) * (1. - t) * points[2][1]
            + (t ** 3.) * points[3][1]
        ),
    )


def cubic_bezier_angle(points, t):
    """ Get the angle of a point along a cubic bezier curve

    Args:
        points (:obj:`list` of :obj:`list` of :obj:`float`): control points
        t (:obj:`float`): position along the curve [0, 1]

    Returns:
        :obj:`float`: angle of the curve at point :obj:`t`
    """
    dx = (
        (1 - t) ** 2 * (points[1][0] - points[0][0])
        + 2 * t * (1 - t) * (points[2][0] - points[1][0])
        + t ** 2 * (points[3][0] - points[2][0])
    )
    dy = (
        (1 - t) ** 2 * (points[1][1] - points[0][1])
        + 2 * t * (1 - t) * (points[2][1] - points[1][1])
        + t ** 2 * (points[3][1] - points[2][1])
    )
    return math.atan2(dy, dx) * 180 / math.pi
